(* -*- mode: tuareg                                                        -*-
 ******************************************************************************
 ** $Id: botmcfg.ml,v 1.26 2003/02/08 03:02:42 albro Exp $
 ******************************************************************************
 **                                 botmcfg.ml
 **
 **              Copyright (C) 2002, UCLA.  All Rights Reserved.
 ******************************************************************************
 ** Authors:
 **      Daniel M. Albro
 ******************************************************************************
 ** Description:
 **      Bottom-up chart parser for mcfgs.
 ******************************************************************************
 ** History:
 **      10/23/02 - [da]: Initial Creation.
 ******************************************************************************
 ** Bugs:
 ******************************************************************************
 ** To Do:
 *****************************************************************************)

type queue_entry = {
  item : Chart.subitem;
  cat : int
}

exception Success of Chart.chart

let is_terminal c = (c = 0)

let terminal_cat = 0

let print_item cat subitem g =
  let length = 
    if cat > 0 then (Array.length subitem) else (Array.length subitem) - 1
  in
    if cat > 0 then
      Printf.printf "%s(%d)" (g#cat_string cat) cat
    else
      Printf.printf "\"%s\"(%d)" (g#terminal_name subitem.(2)) subitem.(2);
    print_string ":[";
    for i = 0 to length - 1 do
      if i <> 0 then
        print_string ", ";
      print_int subitem.(i)
    done;
    print_string "]"

let indirect_sort ary =
  (** Return the indirect sort of an array.  That is, an array of the
    same size whose indices represent the largest item, then the next
    largest, etc. *)
  let al = Array.length ary in
  let indices = Array.init al (function i -> i) in
    for i = 0 to al - 1 do
      for j = i + 1 to al - 1 do
	if ary.(indices.(j)) > ary.(indices.(i)) then
	  let tmp = indices.(i) in
	    indices.(i) <- indices.(j);
	    indices.(j) <- tmp
      done
    done;
    indices

let set_category_information g ch =
  (** Supply the chart /ch/ with information about categories embodied
    in /g/. *)
  (* First, calculate how many times each part of a category gets looked
     up *)
  let ary = Array.init g#num_cats
	      (function i ->
		 if i = 0 then
		   (Array.make 3 0)
		 else
		   (Array.make ((g#arity i) * 2) 0)) in
    for cat = 1 to g#num_cats - 1 do
      let pats = g#patterns cat in
      let process_pattern pat = 
	if pat#multi_cat_pat() then
	  let oc = pat#get_other_cat in
	    for i = 0 to (2 * (g#arity oc)) - 1 do
	      if (pat#get_other_state i) <> -1 then
		ary.(oc).(i) <- ary.(oc).(i) + 1
	    done
      in
	List.iter process_pattern pats
    done;
    (* Now use this information to supply the info for each category *)
    for cat = 1 to g#num_cats - 1 do
      let indices = indirect_sort ary.(cat) in
	ch#set_cat_info cat (g#arity cat) indices
    done

let produce_axioms sentence ch agenda g =
  (** Add all of the items in the sentence into the chart as axioms. *)
  for i = 0 to (Array.length sentence) - 1 do
    let (item : Chart.subitem) = [| i; (i + 1); sentence.(i) |] in
      if ch#add terminal_cat item then begin
        (* print_string "Added ";
           print_item terminal_cat item g;
           print_string "\n"; *)
	Queue.add {item = item; cat = terminal_cat} agenda
      end
  done

let produce_empties ch agenda g sl =
  (** Add items into the chart/agenda for all rules with two empty categories
    on the right hand side. *)
  for cat = 1 to g#num_cats - 1 do
    if g#is_empty_cat cat then begin
      let patterns = g#patterns cat in
      let process_pattern pat =
        if pat#multi_cat_pat() && (g#is_empty_cat pat#get_other_cat) then
          let other_cat = pat#get_other_cat in
          let out_cat = pat#get_out_cat in
          let items0 = Parsetree.lookup_matches cat [|-1;-1|] ch g sl in
          let items1 = List.filter (function it -> pat#check_item it) items0
          in
          let make_new_item item =
            let ms = Parsetree.lookup_matches other_cat 
                       (pat#search_item item) ch g sl in
            let good_ms = List.filter 
                            (function it -> pat#check_other_item it) ms in
            let new_items =
              List.map (function it -> pat#out_item item it) good_ms
            in
            let add_item it =
              if ch#add out_cat it then begin
                (* print_string "Added ";
                   print_item out_cat it g;
                   print_string "\n"; *)
                if out_cat = g#start_cat && it = [|0;sl|] then
                  raise (Success ch)
                else
                  Queue.add {item=it; cat=out_cat} agenda
              end
            in
              List.iter add_item new_items
          in
            List.iter make_new_item items1
      in
        List.iter process_pattern patterns
    end
  done

let produce_terminal_items pat item g ch agenda sl =
  (** Given a rule pattern /pat/ for a terminal rule and an axiomatic
    item /item/, create all items that can be produced by the rule
    described in /pat/ by using /item/.  Add them to the chart
    and the agenda, if they aren't already there. *)
  if item.(2) = pat#get_state 0 then
    let it = [| item.(0); item.(1) |] in
      if ch#add pat#get_out_cat it then begin
        (* print_string "Added ";
           print_item pat#get_out_cat it g;
           print_string "\n"; *)
        if pat#get_out_cat = g#start_cat && it = [|0;sl|] then
          raise (Success ch)
        else
          Queue.add {item=it; cat=pat#get_out_cat} agenda
      end

let produce_nonterminal_items pat item g ch agenda sl =
  (** Given a rule pattern /pat/ for a nonterminal rule and a
    non-axiomatic item /item/, create all items that can be produced
    by the rule described in /pat/ by using /item/.  Add them to 
    the chart and the agenda, if they aren't already there. If a new
    item added to the chart is the success item, signal it with
    Success(ch).
  *)
  if pat#check_item item then
    let new_items =
      if not (pat#multi_cat_pat()) then
	[pat#out_item item [||] ]
      else
        let ms = Parsetree.lookup_matches pat#get_other_cat 
                   (pat#search_item item) ch g sl in
	let good_ms = List.filter (function it -> pat#check_other_item it) ms 
	in
	  List.map (function it -> pat#out_item item it) good_ms
    in
    let add_item it =
      if ch#add pat#get_out_cat it then begin
        (* print_string "Added ";
           print_item pat#get_out_cat it g;
           print_string "\n"; *)
        if pat#get_out_cat = g#start_cat && it = [|0;sl|] then
          raise (Success ch)
        else
          Queue.add {item=it; cat=pat#get_out_cat} agenda
      end
    in
      List.iter add_item new_items

let bottom_up_parse g sentence =
  (** Parse /sentence/ (an array of integers from some symbol table) with
    respect to grammar g.  Return the filled-out chart. *)
  let ch = new Chart.chart g#num_cats in
  let sl = Array.length sentence in
  let agenda = Queue.create() in
    (* Tell the chart how to insert/search on each category in g *)
    set_category_information g ch;
    (* Check for empty sentence *)
    if (Array.length sentence) = 0 then
      if g#is_empty_cat g#start_cat then
        raise (Success ch)
      else
        ch
    else begin
      (* Add the axiomatic items to the chart/agenda *)
      produce_axioms sentence ch agenda g;
      produce_empties ch agenda g sl;
      (* Process the agenda *)
      while not (Queue.is_empty agenda) do
        let entry = Queue.take agenda in
        let patterns = g#patterns entry.cat in
        let process_pattern pat =
          if is_terminal entry.cat then
            produce_terminal_items pat entry.item g ch agenda sl
          else
            produce_nonterminal_items pat entry.item g ch agenda sl
        in
	  List.iter process_pattern patterns
      done;
      ch
    end

let _ =
  let result = ref 0 in
  let parse_sentence g str =
    let words = Str.split (Str.regexp "[ \t]+") str in
    let init_time = Unix.gettimeofday() in
      try
        let snt = Array.of_list (List.map g#terminal_id words) in
          ignore(bottom_up_parse g snt);
          Printf.printf "Failed to parse %s\n" str;
          Printf.printf "Process took %f seconds.\n" (Unix.gettimeofday() -. init_time);
          result := 1
      with
          Success ch -> 
            Printf.printf "Parsed %s\n" str;
            Printf.printf "Process took %f seconds, chart size=%d.\n" 
              (Unix.gettimeofday() -. init_time) ch#chart_size;
            let tree = Parsetree.recover_parse_tree g (List.length words) ch in
              Parsetree.print_parse_tree tree;
              result := 0
        | Grammar.Unknown_terminal wd -> 
            Printf.printf "Failed to parse %s -- \"%s\" is unknown.\n" str wd;
            Printf.printf "Process took %f seconds.\n" 
              (Unix.gettimeofday() -. init_time);
            result := 2
  in
    if (Array.length Sys.argv) >= 2 then
      let gf = Sys.argv.(1) in
      let lexer = Mcfggram.file_lexer gf in
      let g = new Grammar.grammar in
        Mcfggram.parse_grammar g lexer;
        (* print_string "Outputting rules...\n";
        g#output_rules(); *)
        print_string "Checking the grammar...\n";
        g#check_rules();
        print_string "OK.\n";
        print_string "Building pattern database...\n";
        g#build_patterns();
        if (Array.length Sys.argv) = 2 then
          try
            while true do
              print_string "Enter sentence to parse, or hit ^D.\n";
              let ln = read_line () in
                parse_sentence g ln
            done
          with End_of_file -> print_string "Bye!\n"
        else
          for i = 2 to (Array.length Sys.argv) - 1 do
            parse_sentence g Sys.argv.(i)
          done;
        exit(!result)
    else
      begin
        prerr_string "Usage: botmcfg <grammar-file> [<sentence>*]\n";
        exit(-1)
      end
    
    

        

