(* -*- mode: tuareg                                                        -*-
******************************************************************************
** $Id: chart.ml,v 1.7 2003/02/04 08:21:32 albro Exp $
******************************************************************************
**                                  chart.ml
**
**              Copyright (C) 2002, UCLA.  All Rights Reserved.
******************************************************************************
** Authors:
**      Daniel M. Albro
******************************************************************************
** Description:
**      The chart for a bottom-up MCFG chart parser; implementation.  
**      Chart items contain a category and 2n states for some n > 0.
******************************************************************************
** History:
**      11/21/02 - [da]: Initial Creation.
******************************************************************************
** Bugs:
******************************************************************************
** To Do:
*****************************************************************************)

type subitem = int array
(** An item on the chart, as presented to the user (sans cat) *)

(* 
 * We represent a chart as an array of maps from integers (1) to maps
 * from integers (2) to maps from integers (3) to integers (4).  What
 * these integers represent depends on the category.  
 *)

(* Declare integers to be ordered types *)
module OrderedInt =
struct
  type t = int
  let compare x1 x2 = x1 - x2
end

(* Map from integers to some other thing... *)
module AssocInt = Map.Make(OrderedInt)

(* Set of integers *)
module IntSet = Set.Make(OrderedInt)

(* The trie is a map of integers to a map of integers to ... to a set *)
type trie = 
    TrieSet of IntSet.t
  | TrieTree of trie AssocInt.t

(* Exception used internally to chart#add *)
exception Not_added

(***************
 * Chart Class *
 ***************)
class chart nc =
  let order_func i = if i = 0 then [|1;0;2|] else [|1;0|] in
  let tree_depth_func i = if i = 0 then 3 else 2 in
object (self)
  
  (************************
   * Private Data Members *
   ************************)

  val mutable tree = Array.init nc (function i -> TrieTree(AssocInt.empty))
  (** Array from cat to state trie.  The type of this is "trie array" *)
                       
  val num_cats = nc                     
  (** The number of categories in the grammar we're parsing *)

  val tree_depth = Array.init nc tree_depth_func
  (** The depth of the tree for each category.  Will have to set this
    outside somewhere.  Defaults to 2 (arity 1) except for the special
    case of category 0, which has the sort of weird depth 3. *)

  val mutable m_order = Array.init nc order_func
  (** The order to search the trie for each category. In general, for
    maximum efficiency, the order should have unused fields first,
    then fields that get searched on from most frequently searched on
    to least frequently.  It's best to have the field with the largest
    number of possible values last.  Thus unused fields go first,
    since these have only one value (0). *)

  val mutable m_chart_size = 0
  (** Number of items in the chart. *)

  (******************
   * Public Methods *
   ******************)

  method show_cat_info () =
    (** Display the information stored for each category (basically what
      the sort order of the indices for that category is) *)
    for i = 0 to num_cats - 1 do
      Printf.printf "%d) Tree depth: %d, Order: [" i tree_depth.(i);
      for j = 0 to (Array.length m_order.(i)) - 1 do
        if j <> 0 then
          print_string ", ";
        print_int m_order.(i).(j)
      done;
      print_string "]\n"
    done
                        
  method set_cat_info cat arity order =
    (** Set the tree depth and search order for a category.  Assumes
      the category is between 1 and the maximum category. *)
    assert (cat > 0 && cat < num_cats);
    assert (arity > 0 && arity < 10);
    assert ((Array.length order) = (arity * 2));
    tree_depth.(cat) <- arity * 2;
    m_order.(cat) <- order

  method chart_size =
    (** Return the number of items in the chart *)
    m_chart_size

  method contains cat item =
    (** Indicate whether the exact item /cat/:/item/ exists in the chart *)
    assert (cat >= 0 && cat < num_cats);
    let rec trie_partial_lookup subtree level =
      let state = m_order.(cat).(level) in
      let key = item.(state) in
        match subtree with
            TrieSet(set) -> IntSet.mem key set
          | TrieTree(st) ->
              try
                let next = AssocInt.find key st in
                  trie_partial_lookup next (level + 1)
              with Not_found -> false
    in
      trie_partial_lookup tree.(cat) 0

  method matches cat item =
    (** Get a list of all items of category /cat/ that match the
      specification of /item/ -- here, anything matches where a field
      in /item/ is -1, and all other values must match exactly. *)
    assert (cat >= 0 && cat < num_cats);
    let remap lst =
      let ary = Array.make tree_depth.(cat) 0 in
        for i = 0 to tree_depth.(cat) - 1 do
          (* If m_order.(cat).(i) is j then the ith element should be
           * the jth element *)
          ary.(m_order.(cat).(i)) <- List.nth lst i
        done;
	ary
    in
    let rec trie_partial_search subtree level tail =
      assert ((List.length tail) = level);
      let state = m_order.(cat).(level) in
      let key = item.(state) in
      let trie_sub_search skey stree slst =
        let lower = trie_partial_search stree (level + 1) (skey::tail) in
          lower @ slst
      in
        match subtree with
            TrieSet(set) -> 
              if key = -1 then
                List.map (function a -> remap (List.rev (a::tail)))
                  (IntSet.elements set)
              else
                if IntSet.mem key set then
                  [remap (List.rev (key::tail))]
                else
                  []
          | TrieTree(st) ->
              if key = -1 then
                AssocInt.fold trie_sub_search st []
              else
                try
                  let next = AssocInt.find key st in
                    trie_partial_search next (level + 1) (key::tail)
                with Not_found -> []
    in
      trie_partial_search tree.(cat) 0 []

  method add cat item =
    (** Add an item to the chart, if it isn't already there.  If the item is
      actually added to the chart, return true, else return false *)
    let rec trie_insert subtree level =
      assert (level >= 0 && level < tree_depth.(cat));
      let state = m_order.(cat).(level) in
      let key = item.(state) in
        assert (key <> -1);
        match subtree with
            TrieSet(set) ->
              assert (level = (tree_depth.(cat) - 1));
              if IntSet.mem key set then
                raise Not_added
              else
                TrieSet(IntSet.add key set)
          | TrieTree(st) ->
              assert (level < (tree_depth.(cat) - 1));
              let next =
                try
                  AssocInt.find key st
                with Not_found -> 
                  if level < (tree_depth.(cat) - 2) then
                    TrieTree(AssocInt.empty)
                  else
                    TrieSet(IntSet.empty)
              in
              let new_next = trie_insert next (level + 1) in
                TrieTree(AssocInt.add key new_next st)
    in
      try
        tree.(cat) <- (trie_insert tree.(cat) 0);
        m_chart_size <- m_chart_size + 1;
        true
      with Not_added -> false
end


