(* -*- mode: tuareg                                                        -*-
******************************************************************************
** $Id: grammar.ml,v 1.19 2003/02/25 02:38:05 albro Exp $
******************************************************************************
**                                 grammar.ml
**
**              Copyright (C) 2002, UCLA.  All Rights Reserved.
******************************************************************************
** Authors:
**      Daniel M. Albro
******************************************************************************
** Description:
**      Represents an N-Multiple Context Free Grammar in something like
**      Chomsky-Normal Form (rules have one terminal or one or two nonterms.)
**      The main class is "grammar", which is at the bottom of the file.
******************************************************************************
** History:
**      11/21/02 - [da]: Initial Creation.
******************************************************************************
** Bugs:
******************************************************************************
** To Do:  
**  - Simplify the grammar class to remove the distinction between categories
**    and base categories (since those concepts are only useful for grammar
**    intersection/recovery, which we're not doing here).     
*****************************************************************************)

(****************************************************************************
 *                                rule Class                                *
 *--------------------------------------------------------------------------*
 * This consists of a head category, one or two right-hand categories, and  *
 * a map function, here just the index of the map function in the grammar's *
 * list of map functions.                                                   *
 ****************************************************************************)
class rule =
object
  val mutable m_headcat = 0
  val mutable m_rhs = Array.make 2 (-1)
  val mutable m_maptype = 0

  method set_headcat cat = m_headcat <- cat
  method set_rhs i cat = m_rhs.(i) <- cat
  method set_maptype typ = m_maptype <- typ

  method headcat = m_headcat
  method rhs = m_rhs
  method maptype = m_maptype
  method length = if m_rhs.(1) = -1 then 1 else 2
end (* rule *)

(****************************************************************************
 *                       Rule Pattern (rule_pat) Class                      *
 *--------------------------------------------------------------------------*
 * A rule pattern will be stored in a list pertaining to a particular       *
 * category (the "trigger" category).  The rule pattern says that if you    *
 * have a chart item (p1,q1)...(pn,qn):C, you either have automatically an  *
 * item pat#out_states((p1,q1)...(pn,qn)):pat#OC, (in the case of a chain   *
 * rule, if pat#states((p1,q1)...(pn,qn)) is satisfied; or you will have    *
 * such an item if you find an item that matches pat#other_states((p1,q1)...*
 * (pn,qn)) and pat#check_states (on its own states).  The arities array we *
 * pass here is of the form [| <head-cat arity>; <1st rhs arity>; [<2nd rhs *
 * arity>] |].                                                              *
 ****************************************************************************) 
class rule_pat 
  (oc : int)                 (* Output category (category of LHS) *)
  (rn : int)                 (* Rule number in grammar *)
  (num_rhs : int)            (* Number of RHS elements (1 or 2) *)
  arities                    (* Arity array, see above *)
  reversed                   (* true iff binary rule & trigger is 1st rhs el *)
  =
object
  (**************
   * Properties *
   **************)

  (* This array indicates which states of the item we have (the item
     that triggered the use of this pattern) must be identical to
     one another for the rule to trigger.  -1 means that a particular
     state doesn't have to be equivalent to another, any other number
     indicates the other state this state must be equivalent to. *)
  val mutable states = 
    if reversed then
      Array.make (2 * arities.(1)) (-1)
    else
      Array.make (2 * arities.(num_rhs)) (-1)

  (* This array indicates which states of the item we're going to
     look up must be equivalent to states in the item that triggered
     the use of this pattern.  -1 indicates no requirement, any
     other number tells the state we must be equivalent to. *)
  val mutable other_states = 
    if num_rhs = 2 then
      if reversed then
        Array.make (2 * arities.(2)) (-1)
      else
        Array.make (2 * arities.(1)) (-1)
    else
      [||]

  (* This array indicates which states of the item we're going to
     look up must be equivalent to other states in that item. *)
  val mutable check_states = 
    if num_rhs = 2 then
      if reversed then
        Array.make (2 * arities.(2)) (-1)
      else
        Array.make (2 * arities.(1)) (-1)
    else
      [||]

  (* The category we're looking for, if any *)
  val mutable other_cat = 0

  (* The rule number this pattern represents *)
  val mutable rule_num = rn

  (* The category number of the category that will get added to the
     chart if this rule can apply. *)
  val mutable out_cat = oc

  (* Each element of this array is a (n,m) pair signifying that
     the indicated state of the output category should be set from
     the mth state of the nth rhs element. *)
  val mutable out_states = Array.make (2 * arities.(0)) (-1,-1)

  (***********
   * Methods *
   ***********)

  method set_state i v = states.(i) <- v
  method set_other_state i v = other_states.(i) <- v
  method set_check_state i v = check_states.(i) <- v
  method set_other_cat cat = other_cat <- cat
  method set_out_state i v = out_states.(i) <- v

  method get_rule_num = rule_num
  method get_other_cat = other_cat
  method get_out_cat = out_cat
  method multi_cat_pat () = (other_cat <> 0)
  method get_other_state i = other_states.(i)
  method get_state i = states.(i)

  method search_item item =
    (** Given an item, produce the item we need to search for *)
    let ary = Array.make (Array.length other_states) 0 in
      for i = 0 to (Array.length other_states) - 1 do
        if other_states.(i) = -1 then
          ary.(i) <- -1
        else
          ary.(i) <- item.(other_states.(i))
      done;
      ary

  method check_item (item : Chart.subitem) =
    (** Given an item, check to see if it's compatible with this pattern *)
    let result = ref true in
      for i = 0 to (Array.length states) - 1 do
        if states.(i) <> -1 then
          if item.(states.(i)) <> item.(i) then
            result := false
      done;
      !result

  method check_other_item (item : Chart.subitem) =
    (** Given a searched-for item, see if it's compatible with this pattern *)
    let result = ref true in
      for i = 0 to (Array.length check_states) - 1 do
        if check_states.(i) <> -1 then
          if item.(check_states.(i)) <> item.(i) then
            result := false
      done;
      !result

  method out_item (item : Chart.subitem) (other_item : Chart.subitem) =
    (** Given the item representing the rightmost rhs element and the
      other item (use [||] if none), make a new item for the chart *)
    let out_item_value i = 
      match out_states.(i) with
          (0, j) -> if other_cat <> 0 then other_item.(j) else item.(j)
        | (1, j) -> item.(j)
        | _ -> failwith "Illegal value"
    in
      Array.init (Array.length out_states) out_item_value

  method output_self cat =
    print_int out_cat;
    print_string ":[";
    for i = 0 to (Array.length out_states) - 1 do
      if i <> 0 then
        print_string ",";
      print_int (fst out_states.(i));
      print_string ":";
      print_int (snd out_states.(i))
    done;
    print_string "] --> ";
    print_int other_cat;
    print_string ":[";
    for i = 0 to (Array.length other_states) - 1 do
      if i <> 0 then
        print_string ",";
      print_int other_states.(i);
      if Array.length check_states > 0 then begin
        print_string ":";
        print_int check_states.(i)
      end
    done;
    print_string "] ";
    print_int cat;
    print_string ":[";
    for i = 0 to (Array.length states) - 1 do
      if i <> 0 then
        print_string ",";
      print_int states.(i)
    done;
    print_string "]"
end (* rule_pat *)

(****************************************************************************
 *            Rule Reconstruction Pattern (reconstruct_pat) Class           *
 *--------------------------------------------------------------------------*
 * Rule reconstruction pattern (used to reconstruct parse trees; one such   *
 * pattern exists for every rule, more or less).  A reconstruct_pat tells   *
 * what to look for given that the head category of the rule was found in   *
 * the chart to decide whether the referenced rule was used in the          *
 * and if it *was* used, to decide what the elements that combined to form  *
 * the head category were.   It consists of a listing of what states in the *
 * categories on the right hand side of the rule are set from the states in *
 * the head of the rule, plus equivalences between states in the right hand *
 * side.  An equivalence is between state x of rhs item y and state z of    *
 * rhs item w...  We supply as args the number of rhs elements and an array *
 * of arities [| <head arity>; <1st rhs el arity>; [<2nd rhs el arity>] |]. *
 ****************************************************************************)
class reconstruct_pat 
  num_rhs                               (* Number of RHS elements (1 or 2) *)
  arities                               (* Arities of rule elements *)
  =
  let sub_eqv i = Array.make 8 false in
object (self)
  (**************
   * Properties *
   **************)
  val mutable rhs_states = Array.init num_rhs 
                             (function i -> 
                                (Array.make (2 * arities.(i + 1)) (-1)))
  val mutable eqv = []                (* Will contain x00,x02 pairs *)
  val m_arities = arities
                      
  (***********
   * Methods *
   ***********)
  method set_rhs_state i j st = rhs_states.(i).(j) <- st
  method get_rhs_state i j = rhs_states.(i).(j)

  method set_eqv (el1, st1) (el2, st2) =
    if not (self#is_eqv (el1, st1) (el2, st2)) then
      eqv <- ((el1, st1, el2, st2) :: eqv)

  method is_eqv (el1, st1) (el2, st2) = 
    let rec is_eqv_sub (el1, st1) (el2, st2) lst =
      match lst with
          (vel1, vst1, vel2, vst2) :: tl -> 
            if (vel1 = el1 && vst1 = st1 && vel2 = el2 && vst2 = st2)
              || (vel1 = el2 && vst1 = st2 && vel2 = el1 && vst2 = st1) then
                true
            else
              is_eqv_sub (el1, st1) (el2, st2) tl
        | [] -> false
    in
      is_eqv_sub (el1, st1) (el2, st2) eqv

  method fill_pat_unary (pat : rule_pat) =
    (** Fill a bottom-up parsing pattern with all the info in this.
      Assumes the rule is a chain rule. *)
    let handle_eqv = function
        (0, st1, 0, st2) -> pat#set_state st2 st1
      | (_, _, _, _) -> failwith "Should have called fill_pat_binary."
    in
      List.iter handle_eqv eqv;
      for i = 0 to (Array.length rhs_states) - 1 do
        for j = 0 to (Array.length rhs_states.(i)) - 1 do
          if rhs_states.(i).(j) <> -1 then
            pat#set_out_state rhs_states.(i).(j) (i,j)
        done
      done

  method fill_pat_binary (pat1 : rule_pat) (pat2 : rule_pat) =
    (** Fill a bottom-up parsing pattern with all the info in this 
      pat1 is for the right element looking for the left
      element.  pat2 is for the left element looking for the
      right element.
    *)
    
    let handle_eqv = function
        (0, st1, 0, st2) -> 
          pat1#set_check_state st2 st1;
          pat2#set_state st2 st1
      | (0, st1, 1, st2) -> 
          pat1#set_other_state st1 st2;
          pat2#set_other_state st2 st1
      | (1, st1, 0, st2) -> 
          pat1#set_other_state st2 st1;
          pat2#set_other_state st1 st2
      | (1, st1, 1, st2) -> 
          pat1#set_state st2 st1;
          pat2#set_check_state st2 st1
      | _ -> failwith "Illegal equivalence pattern"
    in
    let remapped num =
      if num = 0 then
        1
      else if num = 1 then
        0
      else
        failwith "Non-binary rule."
    in
      List.iter handle_eqv eqv;
      for i = 0 to (Array.length rhs_states) - 1 do
        for j = 0 to (Array.length rhs_states.(i)) - 1 do
          if rhs_states.(i).(j) <> -1 then
            begin
              pat1#set_out_state rhs_states.(i).(j) (i,j);
              pat2#set_out_state rhs_states.(i).(j) (remapped i,j)
            end
        done
      done

  method output_self () =
    print_string "rhs_states:[";
    for i = 0 to (Array.length rhs_states) - 1 do
      print_string "[";
      for j = 0 to (Array.length rhs_states.(i)) - 1 do
        if j <> 0 then
          print_string ", ";
        print_int rhs_states.(i).(j)
      done;
      print_string "]"
    done;
    print_string "], [";
    let print_eqv e =
      match e with
          (e1, st1, e2, st2) -> Printf.printf "(%d,%d)=(%d,%d)," e1 st1 e2 st2
    in
      List.iter print_eqv eqv;
      print_string "]"
end (* reconstruct_pat *)

(****************************************************************************
 *          Lesser Subtypes and Auxilliaries for the Grammar Class          *
 ****************************************************************************)

(* A string with a simple polynomial hash function. *)
module HashedString =
struct
  type t = string
  let equal = (=)
  let hash key =
    let h = ref 0 in
    let l = String.length key in
      for i = 0 to (l - 1) do
        h := (31 * !h) + (int_of_char key.[i])
      done;
      !h
end
  
(* A symbol table that maps strings to 'a *)
module SymbolTable = Hashtbl.Make(HashedString)

(* The type of a mapping function *)
type map_item = int * int
type mapfunc = map_item list array

exception Duplicate_category
exception Category_string of string

let is_terminal r =
  (** Mapping function number 0 indicates that a rule is actually
     a terminal rule. *)
  r#maptype = 0

let fill_rpat rpat (maps : mapfunc) =
  (** [fill_rpat rpat maps] fills out /rpat/, a rule reconstruction pattern,
    with the dependencies reflected in /maps/.  In particular, the rhs_states
    include information about what header states they come from, and the eqv
    list specifies what states within the right hand side are equivalent
    to one another.  Note that /maps/ is an array (one per the arity of head
    category) of lists of (rhs#, arity-idx) pairs. *)
  for i = 0 to (Array.length maps) - 1 do
    (* Within each map, the first element indicates that the "p" value of
       the right-hand side element referenced equals the "p" value of
       the head category for arity i *)
    let prev = ref (List.hd maps.(i)) in
      rpat#set_rhs_state (fst !prev) ((snd !prev) * 2) (i * 2);
      if (List.length maps.(i)) > 1 then
        begin
          let set_eqvs pair =
            rpat#set_eqv (fst !prev, ((snd !prev) * 2) + 1)
              (fst pair, ((snd pair) * 2));
            prev := pair
          in
            List.iter set_eqvs (List.tl maps.(i))
        end;
      (* Now !prev is the last element, and the last element indicates
         something similar to the first *)
      rpat#set_rhs_state (fst !prev) (((snd !prev) * 2) + 1) ((i * 2) + 1)
  done

let fill_pat pat pat2 rpat r =
  (** [fill_pat pat pat2 rpat r binary_rule] fills out a rule parse
    pattern /pat/ according to the information in /r/ and /rpat/. Also 
    fills out pat2 if the rule is binary.
  *)
  match pat2 with
      Some(mirror) ->
        rpat#fill_pat_binary pat mirror;
        pat#set_other_cat r#rhs.(0);
        mirror#set_other_cat r#rhs.(1)
    | None -> 
        rpat#fill_pat_unary pat

let print_mapping map_func =
  (** [print_mapping map_func] outputs a mapping function in human-readable
    form. *)
  let print_mapitem = function
      a, b -> Printf.printf "%d,%d;" a b
  in
    for i = 0 to (Array.length map_func) - 1 do
      print_string "[";
      List.iter print_mapitem map_func.(i);
      print_string "]"
    done

exception Unknown_terminal of string

(****************************************************************************
 *                          Grammar (grammar) Class                         *
 *--------------------------------------------------------------------------*
 * This represents a Multiple Context-Free Grammar and all of the           *
 * information we've managed to wring out of it.  It contains an array of   *
 * rules, a set of category information (category names, distinguished      *
 * start category, category arities, terminal strings, terminal identifiers,*
 * rules headed by each category, information about whether the category is *
 * empty or non-empty, or both), a list of mapping function types, a list   *
 * of patterns used for parsing (instead of looking at the rules directly,  *
 * when parsing we construct these patterns that allow us to detect what    *
 * can be added to the chart in as efficient a manner as possible), and a   *
 * list of patterns used to reconstruct parse trees from charts.            *
 ****************************************************************************)
class grammar =
  let make_rule i = new rule in
  let make_map i = [||] in
object (self)

  (* The array of rules, with size and num. *)
  val mutable m_rules : rule array = Array.init 10 make_rule
  val mutable m_num_rules = 0           (* the number of rules in the G *)
  val mutable m_size_rules = 10         (* the size of the m_rules array *)

  (* The start category *)                               
  val mutable m_start_cat = 1

  (* Symbol table for categories *)
  val mutable m_original_cats = SymbolTable.create 211 (* string->int map *)
  val mutable m_cat_strings = [|"#"|]   (* int->string map *)

  (* Symbol table for terminals *)
  val mutable m_terminals = SymbolTable.create 211 (* string->int map *)
  val mutable m_terminal_strings = [||] (* int->string map *)

  (* Category info *)
  (* Notice that this is a bit more fancy than it needs to be; the extra
     complexity is due to historical factors, i.e. laziness.  In actual
     usage, the "categories" and "base categories" are exactly the same
     thing.  The distinction was due to the original code having 
     "grammar reconstruction," where parsing would create new grammars
     from the original one, but the string names for those categories would
     always be the same as the original names they were derived from. *)
  val mutable m_num_cats = 1            (* Number of categories in G *)
  val mutable m_size_cats = 1           (* Size of the cat_map array *)
  val mutable m_num_base_cats = 1       (* Number of categories *)
  val mutable m_size_base_cats = 1      (* Size of the arity array *)
  val mutable m_arity = [|1|]           (* Arity, by base cat# *)
  val mutable m_cat_map = [|0|]         (* Map from cat# to base cat# *)
  val mutable m_rules_by_cat = [| [] |] (* List of rules for each cat# *)
  val mutable m_empty_cat = [|false|]   (* Category heads empty rules? *)
  val mutable m_nonempty_cat = [|true|] (* Category heads non-empty rules? *)

  (* Map functions -- each map type (where map type is an integer from 0 to
     m_num_map_types) has an associated map function, which is an array
     of lists of (rhs item #, element #) pairs, where (n, m) refers to
     the mth element of the tuple of strings indicated by the nth element
     of the right hand side of a rule. *)
  val mutable m_map_func : mapfunc array = [| [||] |]
  val mutable m_num_map_types = 1       (* Number of map types *)
  val mutable m_sz_map_types = 1        (* Size of the m_map_func array *)

  (* Patterns for parsing.  These tell you all you need to know to see
     if the existence of category C in the chart lets you derive
     category D. These are indexed by category C. *)
  val mutable m_patterns : rule_pat list array = [||]

  (* Patterns for reconstructing parse trees from the chart.  These are
     indexed by rule number.  They tell you what to look for in the
     chart given rule "C -> A B" if you have C in the chart to see if
     you can add (C -> A B) to the parse tree. *)
  val mutable m_reconstructs : reconstruct_pat array = [||]

  (*********
   * Rules *
   *********)

  method get_rule idx = 
    (** [get_rule idx] returns the rule with index /idx/, 0<=idx<numrules *)
    assert (idx >= 0 && idx < m_num_rules);
    m_rules.(idx)

  method add_rule rule =
    (** [add_rule r] adds a new rule to the grammar, increasing numrules *)
    if m_num_rules = m_size_rules then
      ( m_size_rules <- m_size_rules + 10;
        m_rules <- Array.append m_rules (Array.init 10 make_rule) );
    m_rules.(m_num_rules) <- rule;
    m_num_rules <- m_num_rules + 1

  method del_rule idx =
    (** [del_rule idx] deletes the rule at /idx/, 0<=idx<numrules *)
    assert (idx >= 0 && idx < m_num_rules);
    for i = idx to m_num_rules - 2 do
      m_rules.(i) <- m_rules.(i + 1)
    done;
    m_num_rules <- m_num_rules - 1

  method num_rules = 
    (** [num_rules] returns the number of rules in the grammar *)
    m_num_rules

  method rules_headed_by cat =
    (** [rules_headed_by cat] returns a list of rule indices for rules whose
      head category is /cat/ *)
    assert (cat > 0 && cat <= m_num_cats);
    m_rules_by_cat.(cat)

  method output_rule idx =
    (** [output_rule idx] displays a rule *)
    let r = m_rules.(idx) in
      Printf.printf "%s(%d) --> " (self#cat_string r#headcat) r#headcat;
      if r#maptype = 0 then
        Printf.printf "\"%s\"(%d)" (self#terminal_name r#rhs.(0)) r#rhs.(0)
      else
        begin
          Printf.printf "%s(%d)" (self#cat_string r#rhs.(0)) r#rhs.(0);
          if r#length > 1 then
            Printf.printf " %s(%d)" (self#cat_string r#rhs.(1)) r#rhs.(1);
          print_string " ";
          print_mapping (self#mapping_function r#maptype)
        end

  method output_rules () =
    (** [output_rules ()] displays the list of rules in the grammar *)
    for i = 0 to m_num_rules - 1 do
      self#output_rule i;
      print_string "\n"
    done

  method check_rules () =
    (** [check_rules ()] makes sure category arities and that sort of thing
      are consistent. *)
    let check_rule ridx = 
      let r = m_rules.(ridx) in
      let check_terminal_rule () =
        if (self#arity r#headcat) <> 1 then
          begin
            print_string "Offending rule: ";
            self#output_rule ridx;
            print_string "\n";
            failwith "Terminal rule with arity != 1"
          end
      in
      let check_nonterminal_rule () =
        let mf = self#mapping_function r#maptype in
        let check_mapitem mi =
          if (snd mi) >= self#arity r#rhs.(fst mi) then
            begin
              print_string "Offending rule: ";
              self#output_rule ridx;
              print_string "\n";
              failwith "Mapping function does not match rhs arities."
            end
        in
          for i = 0 to (Array.length mf) - 1 do
            List.iter check_mapitem mf.(i)
          done
      in
        if r#maptype = 0 then
          check_terminal_rule()
        else
          check_nonterminal_rule()
    in
      for i = 0 to m_num_rules - 1 do
        check_rule i
      done
        
  (**************
   * Categories *
   **************)

  method start_cat =
    (** [start_cat] returns the distinguished start category for the grammar *)
    m_start_cat

  method set_start_cat cat =
    (** [set_start_cat cat] sets the start category to /cat/,
      1<=cat<num_cats *)
    assert (cat > 0 && cat < m_num_cats);
    m_start_cat <- cat

  method new_category bc =
    (** [new_category bc] Adds a new category to the grammar whose
      base category is /bc/ (1<=bc<num_base_cats).  Returns the
      category number *)
    assert ( bc > 0 && bc < m_num_base_cats );
    if m_num_cats = m_size_cats then
      begin
        let num_new_cats = ((m_num_cats + 10) - m_size_cats) in
          m_cat_map <- Array.append m_cat_map (Array.make num_new_cats 0);
          m_empty_cat <- Array.append m_empty_cat 
            (Array.make num_new_cats false);
          m_nonempty_cat <- Array.append m_nonempty_cat 
            (Array.make num_new_cats false);
          m_size_cats <- m_size_cats + num_new_cats
      end;
    m_cat_map.(m_num_cats) <- bc;
    m_num_cats <- m_num_cats + 1;
    m_num_cats - 1

  method base_cat cat =
    (** [base_cat cat] returns the base category of /cat/,
      0<=cat<m_num_cats *)
    assert (cat >= 0 && cat < m_num_cats);
    m_cat_map.(cat)

  method lookup_base_cat str =
    (** [lookup_base_cat str] returns the base category with string
      value /str/ or raises Not_found *)
    SymbolTable.find m_original_cats str

  method arity cat =
    (** [arity cat] returns the arity of category /cat/,
      0<=cat<num_cats *)
    assert (cat >= 0 && cat < m_num_cats);
    m_arity.(m_cat_map.(cat))

  method num_cats =
    (** [num_cats] is the number of categories in the grammar *)
    m_num_cats

  method new_base_category name =
    (** [new_base_category name] creates a new base category labeled
      as string /name/ and returns its base category number.  If the
      string has already been used as a base category, returns its
      base category number. *)
    if SymbolTable.mem m_original_cats name then
      SymbolTable.find m_original_cats name
    else
      begin
        m_num_base_cats <- m_num_base_cats + 1;
        SymbolTable.add m_original_cats name (m_num_base_cats - 1);
        m_cat_strings <- Array.append m_cat_strings [|name|];
        if m_num_base_cats > m_size_base_cats then begin
          m_size_base_cats <- m_size_base_cats + 10;
          m_arity <- Array.append m_arity (Array.make 10 0)
        end;
        self#new_category (m_num_base_cats - 1)
      end

  method set_arity cat new_arity =
    (** [set_arity cat new_arity] sets the arity for /cat/ to /new_arity/,
      1<=new_arity<= 2.  If the arity was already set to something else,
      raises Duplicate_category exception. *)
    assert (cat > 0 && cat < m_num_cats);
    assert (new_arity > 0);
    if m_arity.(m_cat_map.(cat)) = 0 then
      m_arity.(m_cat_map.(cat)) <- new_arity
    else if m_arity.(m_cat_map.(cat)) <> new_arity then
      raise Duplicate_category

  method num_base_cats =
    (** [num_base_cats] is the number of base categories *)
    m_num_base_cats

  method cat_string cat =
    (** [cat_string cat] returns the string corresponding to a category. *)
    assert (cat > 0 && cat < m_num_cats);
    let bc = m_cat_map.(cat) in
      m_cat_strings.(bc)
        
  method is_empty_cat cat =
    (** [is_empty_cat cat] tells whether the category heads any
      productions of the form 'C --> ""' *)
    assert (cat >= 0 && cat < m_num_cats);
    m_empty_cat.(cat)

  method is_nonempty_cat cat =
    (** [is_nonempty_cat cat] tells whether the category heads any
      productions other than of the form 'C --> ""' *)
    assert (cat >= 0 && cat < m_num_cats);
    m_nonempty_cat.(cat)

  method set_empty_cat cat value =
    (** [set_empty_cat cat value] tells whether the category heads any
      productions of the form 'C --> ""' *)
    assert (cat > 0 && cat < m_num_cats);
    m_empty_cat.(cat) <- value

  method set_nonempty_cat cat value =
    (** [set_nonempty_cat cat] tells whether the category heads any
      productions other than of the form 'C --> ""' *)
    assert (cat > 0 && cat < m_num_cats);
    m_nonempty_cat.(cat) <- value

  (*************
   * Terminals *
   *************)
  method new_terminal str =
    (** [new_terminal str] adds /str/ as a new terminal item if it isn't
      already one.  In either case, we return the terminal's ID number... 
      If str is "", returns -1 *)
    if str = "" then
      -1
    else if SymbolTable.mem m_terminals str then
      SymbolTable.find m_terminals str
    else
      let term_id = Array.length m_terminal_strings in
        SymbolTable.add m_terminals str term_id;
        m_terminal_strings <- Array.append m_terminal_strings [|str|];
        term_id
          
  method terminal_id name =
    (** [terminal_id name] gets the terminal id number corresponding
      to a string. Raises Unknown_terminal if there isn't such a
      terminal known to the grammar. *)
    if name = "" then
      -1
    else
      try
        SymbolTable.find m_terminals name
      with Not_found -> raise (Unknown_terminal name)

  method terminal_name id =
    (** [terminal_name id] gets the string form of a terminal from its integer
      form, or raises Not_found *)
    if id >= 0 && id < (Array.length m_terminal_strings) then
      m_terminal_strings.(id)
    else if id = -1 then
      ""
    else
      raise Not_found

  (************
   * Patterns *
   ************)

  method private clear_patterns () =
    (** Reset the patterns to a default state *)
    let new_reconstruct i =
      let r = m_rules.(i) in
      let num_rhs = r#length in
      let arities = Array.init (num_rhs + 1)
                      (function i ->
                         if i = 0 then
                           self#arity r#headcat
                         else if r#maptype = 0 then
                           1
                         else
                           self#arity r#rhs.(i - 1) 
                      )
      in
        new reconstruct_pat num_rhs arities
    in
      m_patterns <- Array.init m_num_cats (function i -> []);
      m_reconstructs <- Array.init m_num_rules new_reconstruct;
      m_rules_by_cat <- Array.init m_num_cats (function i -> [])

  method private calculate_empty_cats () =
    (** Calculate which categories count as empty categories and
      which count as nonempty.  These sets are not disjoint, 
      necessarily. *)
    let changed = ref false in
    let propagate_chain r = 
      if self#is_empty_cat r#rhs.(0) then
        begin
          if not (self#is_empty_cat r#headcat) then
            changed := true;
          self#set_empty_cat r#headcat true
        end;
      if self#is_nonempty_cat r#rhs.(0) then
        begin
          if not (self#is_nonempty_cat r#headcat) then
            changed := true;
          self#set_nonempty_cat r#headcat true
        end
    in
      for i = 0 to m_num_rules - 1 do
        (* Basically we're propagating empty- and nonempty-hood from
           terminal rules upward via chain rules. *)
        let r = m_rules.(i) in
          if r#maptype <> 0 then
            if r#length = 1 then
              propagate_chain r
            else
              begin
                if not (self#is_nonempty_cat r#headcat) then
                  changed := true;
                self#set_nonempty_cat r#headcat true
              end
      done;
      if !changed then
        (* Continue until information percolates to the top of the grammar *)
        self#calculate_empty_cats()

  method private display_empty_cats () =
    for i = 1 to m_num_cats - 1 do
      if self#is_empty_cat i then
        Printf.printf "%s(%d) is an empty category.\n" (self#cat_string i) i;
      if self#is_nonempty_cat i then
        Printf.printf "%s(%d) is a non-empty category.\n" (self#cat_string i) i
    done

  method build_patterns () =
    (** [build_patterns ()] gathers information about the grammar for use
      by parsing *)
    self#clear_patterns();
    self#calculate_empty_cats();
    (* self#display_empty_cats(); *)
    for i = 0 to m_num_rules - 1 do
      let r = m_rules.(i) in
      let multi_cat = r#length > 1 in
      let multi_map = (self#arity(r#headcat) > 1) in
      let num_rhs = r#length in
      let arities = Array.init (num_rhs + 1)
                      (function j ->
                         if j = 0 then
                           self#arity r#headcat
                         else if is_terminal r then 
                           1 
                         else 
                           self#arity r#rhs.(j - 1) 
                      )
      in
      let pat = new rule_pat r#headcat i num_rhs arities false in
      let pat2 = 
        if multi_cat then 
          Some(new rule_pat r#headcat i num_rhs arities true)
        else
          None
      in
      let pat_cat =
        if is_terminal r then
          begin assert (not multi_cat && not multi_map);
            (* Create a pattern that looks for nothing to build the NT *)
            pat#set_state 0 r#rhs.(0);
            0
          end
        else
          let maps = m_map_func.(r#maptype) in
            fill_rpat m_reconstructs.(i) maps;
            fill_pat pat pat2 m_reconstructs.(i) r;
            if multi_cat then
              r#rhs.(1)
            else
              r#rhs.(0)
      in
        (* We don't need the patterns for empty rules, or for triggers
           that won't happen, i.e. aren't non-empty categories, (unless
           the mirror trigger exists and won't happen either. *)
        if (not (is_terminal r)) || r#rhs.(0) <> -1 then
          if (self#is_nonempty_cat pat_cat) 
            || (multi_cat && (not (self#is_nonempty_cat r#rhs.(0)))) then
              begin
                (* print_string "Adding pattern ";
                pat#output_self pat_cat;
                print_string "\n"; *)
                m_patterns.(pat_cat) <- (pat :: m_patterns.(pat_cat))
              end;
        m_rules_by_cat.(r#headcat) <- (i :: m_rules_by_cat.(r#headcat));
        match pat2 with
            Some(mirror) ->
              if self#is_nonempty_cat r#rhs.(0) then
                begin
                  (* print_string "   and pattern ";
                  mirror#output_self r#rhs.(0);
                  print_string "\n"; *)
                  m_patterns.(r#rhs.(0)) <- (mirror :: m_patterns.(r#rhs.(0)))
                end
          | None -> ()
    done

  method patterns cat =
    (** [patterns cat] returns the patterns that tell you what categories
      to look up from cat *)
    assert (cat >= 0 && cat < m_num_cats);
    m_patterns.(cat)

  method reconstruction_pattern ridx =
    (** [reconstruction_pattern ridx] returns the rule reconstruction pattern
      for rule ridx *)
    assert (ridx >= 0 && ridx < m_num_rules);
    m_reconstructs.(ridx)

  (*********************
   * Mapping Functions *
   *********************)

  method add_mapping_func map_func =
    (** [add_mapping_func map_func] adds a new mapping function
      and returns its index.  If the same function was already there,
      returns the index it had *)
    let result = ref (-1) in
      for i = 1 to m_num_map_types - 1 do
	if m_map_func.(i) = map_func then
	  result := i
      done;
      if !result <> (-1) then
	!result
      else
	begin
          if m_num_map_types = m_sz_map_types then begin
	    m_sz_map_types <- m_sz_map_types + 10;
	    m_map_func <- Array.append m_map_func (Array.init 10 make_map)
          end;
	  m_map_func.(m_num_map_types) <- map_func;
	  m_num_map_types <- m_num_map_types + 1;
	  (m_num_map_types - 1)
        end

  method mapping_function midx =
    (** [mapping_function midx] get mapping function number /midx/ *)
    assert (midx > 0 && midx < m_num_map_types);
    m_map_func.(midx)

  method arity_of_type midx =
  (** 
    @param midx Mapping function index
    @return the arity of the head category of any rule using map type [midx] *)
    assert (midx >= 0 && midx < m_num_map_types);
    if midx = 0 then
      1
    else
      Array.length m_map_func.(midx)
end (* grammar *)
