(* -*- mode: tuareg                                                        -*-
******************************************************************************
** $Id: grammar.mli,v 1.14 2003/02/25 02:37:35 albro Exp $
******************************************************************************
**                                grammar.mli
**
**              Copyright (C) 2002, UCLA.  All Rights Reserved.
******************************************************************************)
(**
  Represents an N-Multiple Context Free Grammar in something like
  Chomsky-Normal Form (rules have one terminal or one or two nonterms.)

  @author Daniel M. Albro
  @version $Revision: 1.14 $
  @since 11/21/02 - [da]: Initial Creation.
*)

(****************************************************************************
 *                                rule Class                                *
 ****************************************************************************)

(** This consists of a head category, one or two right-hand categories, and
  a map function, here just the index of the map function in the grammar's
  list of map functions. *)
class rule :
object
  method set_headcat : int -> unit     
    (** Sets the head category to {i cat}
      @param cat Head category
      @return nothing
    *)

  method set_rhs : int -> int -> unit  
    (** Sets the {i i}th RHS element to {i cat} 
      @param i Right hand side element to set
      @param cat Category to put at rhs# idx.
      @return nothing
    *)

  method set_maptype : int -> unit
    (** Notes that the mapping function is indexed as
      number {i typ} in the grammar this rule is stored in. 
      @param typ Map function type of this rule.
      @return nothing
    *)

  method headcat : int        (** @return the head category *)
  method rhs : int array      (** @return array of RHS NT cats or T ids *)
  method maptype : int        (** @return mapping function type (0 for T) *)
  method length : int         (** @return length of rhs array *)
end

(****************************************************************************
 *                       Rule Pattern (rule_pat) Class                      *
 ****************************************************************************)

(** A rule pattern will be stored in a list pertaining to a particular
  category (the "trigger" category).  The rule pattern says that if you
  have a chart item (p1,q1)...(pn,qn):C, you either have automatically
  an item pat#out_states((p1,q1)...(pn,qn)):pat#OC, (in the case of a
  chain rule, if pat#states((p1,q1)...(pn,qn)) is satisfied; or you
  will have such an item if you find an item that matches
  pat#other_states((p1,q1)...  (pn,qn)) and pat#check_states (on its
  own states).  The arities array we pass here is of the form [[|
  <head-cat arity>; <1st rhs arity>; [<2nd rhs arity>] |]].  
  @param oc Output category (LHS cat)
  @param rn Rule number in grammar
  @param num_rhs Number of RHS elements (1 or 2)
  @param arities Arity array, see above
  @param reversed true iff binary rule and the trigger is the 1st element
*)
class rule_pat : 
  int ->       
  int ->       
  int ->       
  int array -> 
  bool ->      
object
  method set_state : int -> int -> unit
    (** Sets state number {i i} to value {i v}.  Here a state
      number is a specifier for what the item that triggers this pattern
      should look like, so [set_state i v] indicates that state {i i} of the
      trigger should match state {i v} of the trigger. 
      @param i Index of state to set.
      @param v Value to set.
      @return nothing
    *)

  method set_other_state : int -> int -> unit
    (** Sets "other state" number {i i} to value {i v}.
      This signifies that the {i i}th state of the category that we're going
      to look for in the chart as a result of triggering this pattern should
      match the {i v}th state of the trigger item. 
      @param i Index of state to set.
      @param v Value to set the state to.
      @return nothing
    *)

  method set_check_state : int -> int -> unit
    (** Sets "check state" number {i i} to value {i v}.
      This means that the {i i}th state of the item we looked for in the 
      chart (the so-called "other" item) must match the {i v}th state of
      that same item. 
      @param i Index of check state to set
      @param v Value to set the ith check state to.
      @return nothing
    *)

  method set_other_cat : int -> unit
    (** Says that the presence of the trigger item for
      this pattern indicates that we should look for an item in the chart
      of category {i cat}.  This defaults to 0, indicating that the pattern
      is for a chain rule. 
      @param cat Non-trigger category number.
      @return nothing
    *)

  method set_out_state : int -> int * int -> unit
    (** Indicates that if the pattern
      matches, the {i i}th state of the item to be produced will come from
      the {i st-idx}th state of the {i el-idx}th right-hand-side element. 
      @param i Out-state number to specify.
      @param v Value to set out-state i to (a tuple: {i (el-idx, st-idx)})
      @return nothing
    *)
   
  method get_rule_num : int  (** @return the rule number this pattern is for *)
  method get_other_cat : int (** @return category to look up in the chart *)
  method get_out_cat : int   (** @return category of item to be produced *)
  method multi_cat_pat : unit -> bool (** @return whether |RHS| > 1 *)

  method get_other_state : int -> int   
    (** @return the trigger item state that must be matched by the item
      to be searched for in the chart.  This value can be -1 to
      indicate that there is no match requirement.
      @param i index of non-trigger-item state specification to look at
    *)

  method get_state : int -> int
    (** @param i index of trigger state to check.
      @return State number of trigger item, other than {i i}, that 
              state number {i i} of trigger item must be equal to, or
              -1 if there is no requirement.
    *)

  method search_item : Chart.subitem -> Chart.subitem
    (** @param item Item that triggered the use of this pattern.
        @return the search key needed to look in the chart
                and see if the rule indicated by this pattern can 
                apply. 
    *)

  method check_item : Chart.subitem -> bool
    (** @param item Item that triggered the use of this pattern.
        @return whether the states in {i item} meet the self-identity
                requirements of the pattern (requirements that certain 
                states in the item be identical to other states in the 
                item). *)

  method check_other_item : Chart.subitem -> bool
    (** @param item An item found in the chart using the key produced by
                    [search_item]
        @return whether the states in {i item} meet the requirements imposed
                by the pattern (the requirement that certain states in the
                item be identical to other states in that same item). *)

  method out_item : Chart.subitem -> Chart.subitem -> Chart.subitem
    (** @param item Item that triggered the use of this pattern.
        @param other_item Other item involved in the rule, or [[||]] if the
                          rule is a chain rule
        @return the item that is produced by application of the rule. *)
    
  method output_self : int -> unit
    (** Shows the pattern in human-readable form.
        @param cat Category of the trigger item for this pattern.
        @return nothing *)
end (* rule_pat *)

(****************************************************************************
  *            Rule Reconstruction Pattern (reconstruct_pat) Class           *
****************************************************************************)

(** Rule reconstruction pattern (used to reconstruct parse trees; one
  such pattern exists for every rule, more or less).  A
  reconstruct_pat tells what to look for given that the head category
  of the rule was found in the chart (the found item will be referred
  to in the comments below as the "left-hand-side item") to decide
  whether the referenced rule was used in the and if it *was* used, to
  decide what the elements that combined to form the head category
  were.  It consists of a listing of what states in the categories on
  the right hand side of the rule are set from the states in the head
  of the rule, plus equivalences between states in the right hand
  side.  An equivalence is between state x of rhs item y and state z
  of rhs item w...  We supply as args the number of rhs elements and
  an array of arities [[| <head arity>; <1st rhs el arity>; [<2nd rhs
  el arity>] |]].

  @param num_rhs Number of RHS elements (1 or 2)
  @param arities Arities of rule elements (see above)
*)
class reconstruct_pat : 
  int ->               
  int array ->         
object
  method set_rhs_state : int -> int -> int -> unit
    (** Indicates that the {i j}th state of the
      {i i}th element of the right hand side comes from the {i st}th
      state in the left-hand-side element 
      @param i Indicated element of the right hand side of the rule
      @param j Indicated state of the indicated element
      @param st Output item state the right-hand-side state comes from or,
                alternatively, produces
      @return nothing
    *)

  method get_rhs_state : int -> int -> int
    (** 
      @return which of the left-hand-side element's states is the origin
              of the {i j}th state of the {i i}th element of the right
              hand side.  The value is -1 if that state is set by equivalence
              relations rather than directly from the left-hand-side element.
      @param i Indicated element of the right hand side of the rule
      @param j Indicated state of the indicated element *)

  method set_eqv : int * int -> int * int -> unit
    (** Modifies this pattern to note that the indicated RHS element states
      are equivalent, that is, are required to be identical.
      @param el1 First indicated RHS element
      @param st1 Indicated state of 1st indicated RHS element
      @param el2 Second indicated RHS element
      @param st2 Indicated state of the 2nd indicated RHS element
      @return nothing *)

  method is_eqv : int * int -> int * int -> bool
    (** @param el1 First indicated RHS element
      @param st1 Indicated state of 1st indicated RHS element
      @param el2 Second indicated RHS element
      @param st2 Indicated state of the 2nd indicated RHS element
      @return whether the indicated states are required to match. *)

  method fill_pat_unary : rule_pat -> unit
    (** Fills in {i pat} with as much information as possible, using the
      information stored in {i this}.  This will only be useful to the
      extent that this reconstruction pattern has already been filled with
      information!  This method is specific to chain rules.
      @param pat Pattern to be used in parsing for the rule this pattern
                   will be reconstructing after parsing is done.
      @return nothing *)

  method fill_pat_binary : rule_pat -> rule_pat -> unit
    (** Fills in {i pat1} and {i pat2} with as much information as
      possible, using the information stored in {i this}.  This will
      only be useful to the extent that this reconstruction pattern
      has already been filled with information!  This method is
      specific to binary rules.

      @param pat1 Pattern to be used in parsing for the rule this pattern
                  will be reconstructing after parsing is done.  This pattern
                  is the one that represents the rightmost RHS element as
                  a triggering item.
      @param pat2 Pattern representing the leftmost RHS element as a triggering
                  item.
      @return nothing *)

  method output_self : unit -> unit
    (** Shows the reconstruction pattern in human-readable format. 
      @return nothing *)
end

(* Types for representing a map function *)

type map_item = int * int               (** Pair: RHS index, element index *)
type mapfunc = map_item list array      (** String map function for MCFGs *)

val is_terminal : rule -> bool
  (** @param r Rule being checked 
    @return whether rule {e r} is a terminal one or not, that is, 
            whether the right hand side is a terminal element. *)

exception Duplicate_category
  (** Raised when a grammar being built has two different arities for the same
    category. *)

exception Unknown_terminal of string
(** Raised if a {i terminal_id} is looked up for a terminal that doesn't
  exist.  This typically happens when trying to parse a sentence with some
  vocabulary not mentioned in the grammar. *)

(****************************************************************************
 *                          Grammar (grammar) Class                         *
 ****************************************************************************)

(** This represents a Multiple Context-Free Grammar and all of the
  information we've managed to wring out of it.  It contains an array of
  rules, a set of category information (category names, distinguished
  start category, category arities, terminal strings, terminal
  identifiers, rules headed by each category, information about whether
  the category is empty or non-empty, or both), a list of mapping
  function types, a list of patterns used for parsing (instead of
  looking at the rules directly, when parsing we construct these
  patterns that allow us to detect what can be added to the chart in as
  efficient a manner as possible), and a list of patterns used to
  reconstruct parse trees from charts. *)
class grammar :
object
  (*********
   * Rules *
   *********)

  method get_rule : int -> rule
  (** @param idx Index of rule to get
    @return the rule with index {i idx}, 0<=idx<num_rules *)

  method add_rule : rule -> unit
  (** Adds a new rule to the grammar.
    @param rule Rule to add
    @return nothing *)

  method del_rule : int -> unit
  (** Deletes the rule at {i idx}.
    @param idx Index of rule to delete, 0 <= {i idx} <= num_rules
    @return nothing
  *)

  method num_rules : int
  (** @return the number of rules in the grammar *)

  method rules_headed_by : int -> int list
  (** @param cat Head category
    @return a list of rule indices for rules whose head category is {i cat} *)

  method output_rule : int -> unit
    (** Displays a rule to stdout.
      @param idx Index of rule to output (0 <= {i idx} <= num_rules) 
      @return nothing *)

  method output_rules : unit -> unit
    (** Displays the list of rules in the grammar 
      @return nothing *)

  method check_rules : unit -> unit
    (** Makes sure category arities and that sort of thing are consistent. 
      @return nothing *)

  (**************
   * Categories *
   **************)

  method start_cat : int
  (** @return the distinguished start category for the grammar *)

  method set_start_cat : int -> unit
  (** Sets the start category to {i cat}
    @param cat Start category for the grammar, 1 <= {i cat} < num_cats 
    @return nothing *)

  method new_category : int -> int
  (** Adds a new category to the grammar whose base category is {i bc}.
    @param bc Number of base category for new cat (1 <= {i bc} < num_base_cats)
    @return the category number *)

  method base_cat : int -> int
  (** @return the base category of {i cat}
    @param cat Category whose base category we want, 0 <= {i cat} < num_cats *)

  method lookup_base_cat : string -> int
  (** @return the base category with string value {i str}
    @param str String to look up
    @raise Not_found if no base category is named {i str} *)

  method arity : int -> int
  (** @return the arity of category {i cat}
      @param cat Category whose arity we want to know, 
                 0 <= [cat] < [num_cats] *)

  method num_cats : int
  (** @return the number of categories in the grammar *)

  method new_base_category : string -> int
  (** If there is no base category already named {i name}, creates a
    new base category with that name.
    
    @param name Name of base category to add or look up
    @return the (possibly newly assigned) base category number of the
    named category *)

  method set_arity : int -> int -> unit
    (** Sets the arity for [cat] to [new_arity].

      @param cat Category whose arity we wish to set
      @param new_arity Arity (dimension) we want to set the category to have
       (1 <= [new_arity] <= 2)
      @return nothing
      @raise Duplicate_category if [cat] already had an assigned arity
      different from [new_arity]. *)

  method num_base_cats : int
    (** @return the number of base categories *)

  method cat_string : int -> string
    (** @return the string corresponding to [cat]
      @param cat Category to look up, 1 <= [cat] <= [num_cats] *)

  method is_empty_cat : int -> bool
    (** @return whether the category heads any productions of the 
                form 'C --> ""' 
      @param cat Category to check
    *)

  method is_nonempty_cat : int -> bool
    (** 
      @param cat Category to check
      @return whether the category heads any productions other than ones 
              of the form 'C --> ""' 
    *)

  method set_empty_cat : int -> bool -> unit
    (** Indicate whether [cat] heads any empty productions.
      @param cat Category in question
      @param value Whether [cat] heads a production like 'C --> ""'
      @return nothing *)

  method set_nonempty_cat : int -> bool -> unit
    (** Indicate whether [cat] heads any non-empty productions.
      @param cat Category in question
      @param value Whether [cat] heads any production {i not} of the form
                   'C --> ""'
      @return nothing *)

  (*************
   * Terminals *
   *************)
  method new_terminal : string -> int
    (** Adds [str] as a new terminal item if it isn't already one.
      @param str Terminal to add
      @return ID number of the terminal (whether just added or not) *)

  method terminal_id : string -> int
    (**
      @param name Name of terminal whose ID we want
      @return the terminal id number corresponding to [name]
      @raise Unknown_terminal(name) if there is no known terminal with the
             string form [name] *)

  method terminal_name : int -> string
    (** 
      @param id Numerical ID of terminal whose name we want to know
      @return String representing the terminal whose ID is [id]
      @raise Not_found if no terminal has the ID [id] *)

  (************
   * Patterns *
   ************)

  method build_patterns : unit -> unit
  (** Gathers information about the grammar for use by parsing 
    @return nothing *)

  method patterns : int -> rule_pat list
  (** 
    @param cat Category of a trigger item
    @return the patterns that tell you what categories to look up from [cat]
            to see what rules can apply to add new items to the chart *)

  method reconstruction_pattern : int -> reconstruct_pat
  (** 
    @param ridx Index of rule whose reconstruction pattern is desired
    @return the rule reconstruction pattern for rule [ridx] *)

  (*********************
   * Mapping Functions *
   *********************)

  method add_mapping_func : mapfunc -> int
  (** Adds [map_func] function to the grammar, if it wasn't already there.
    @param map_func String mapping function to add
    @return the index of the mapping function (whether just added or not) *)

  method mapping_function : int -> mapfunc
  (**
    @param midx Index of mapping function desired
    @return the mapping function with index [midx] *)

  method arity_of_type : int -> int
  (** 
    @param midx Mapping function index
    @return the arity of the head category of any rule using map type [midx] *)
end (* grammar *)
