(* -*- mode: tuareg                                                        -*-
******************************************************************************
** $Id: mcfggrammar.ml,v 1.7 2003/01/26 03:40:03 albro Exp $
******************************************************************************
**                               mcfggrammar.ml
**
**              Copyright (C) 2002, UCLA.  All Rights Reserved.
******************************************************************************
** Authors:
**      Daniel M. Albro
******************************************************************************
** Description:
**      Stream parser for MCFG grammars.
******************************************************************************
** History:
**      12/03/02 - [da]: Initial Creation.
******************************************************************************
** Bugs:
******************************************************************************
** To Do:
*****************************************************************************)

(* Lexer from a string *)
let base_lexer stream =
  Genlex.make_lexer ["-->"; "["; "]"; "("; ")"; ";"; ","] stream

let lexer s = base_lexer (Stream.of_string s)

let file_lexer filename =
  let fch = open_in filename in base_lexer (Stream.of_channel fch)

(* The BNF for this is as follows:

   rule ::= ID "-->" rhs
   rhs  ::= STRING | ID rhs2
   rhs2 ::= ID maps | maps
   maps ::= "[" map "]" maps | /*empty*/
   map  ::= INT "," INT map2
   map2 ::= ";" map | /*empty*/
*)   

let rec parse_grammar g s =
  try
    let (strm__ : _ Stream.t) = s in
    match Stream.peek strm__ with
      Some (Genlex.Ident x) ->
        Stream.junk strm__;
        begin match Stream.peek strm__ with
          Some (Genlex.Kwd "-->") ->
            Stream.junk strm__;
            let r =
              try parse_rhs g strm__ with
                Stream.Failure -> raise (Stream.Error "")
            in
            let u =
              try parse_grammar g strm__ with
                Stream.Failure -> raise (Stream.Error "Error 1")
            in
            r#set_headcat (g#new_base_category x);
            if r#maptype = 0 then g#set_arity r#headcat 1
            else g#set_arity r#headcat (g#arity_of_type r#maptype);
            if x = "S" then g#set_start_cat r#headcat;
            if r#maptype = 0 then
              if r#rhs.(0) = -1 then g#set_empty_cat r#headcat true
              else g#set_nonempty_cat r#headcat true;
            g#add_rule r
        | _ -> raise (Stream.Error "")
        end
    | _ -> ()
  with
    Stream.Error _ -> failwith "Stream error in parse_grammar"
and parse_rule g s =
  try
    let (strm__ : _ Stream.t) = s in
    match Stream.peek strm__ with
      Some (Genlex.Ident x) ->
        Stream.junk strm__;
        begin match Stream.peek strm__ with
          Some (Genlex.Kwd "-->") ->
            Stream.junk strm__;
            let r =
              try parse_rhs g strm__ with
                Stream.Failure -> raise (Stream.Error "")
            in
            r#set_headcat (g#new_base_category x);
            if r#maptype = 0 then g#set_arity r#headcat 1
            else g#set_arity r#headcat (g#arity_of_type r#maptype);
            if x = "S" then g#set_start_cat r#headcat;
            if r#maptype = 0 then
              if r#rhs.(0) = -1 then g#set_empty_cat r#headcat true
              else g#set_nonempty_cat r#headcat true;
            r
        | _ -> raise (Stream.Error "")
        end
    | _ -> failwith "Parse error in parse_rule"
  with
    Stream.Error _ -> failwith "Stream error in parse_rule"
and parse_rhs g s =
  try
    let (strm__ : _ Stream.t) = s in
    match Stream.peek strm__ with
      Some (Genlex.String term) ->
        Stream.junk strm__;
        let x = g#new_terminal term in
        let r = new Grammar.rule in r#set_rhs 0 x; r#set_maptype 0; r
    | Some (Genlex.Ident x1) ->
        Stream.junk strm__;
        let r1 =
          try parse_rhs2 g strm__ with
            Stream.Failure -> raise (Stream.Error "Error 3")
        in
        r1#set_rhs 0 (g#new_base_category x1); r1
    | _ -> failwith "Parse error in parse_rhs"
  with
    Stream.Error _ -> failwith "Stream error in parse_rhs"
and parse_rhs2 g s =
  try
    let (strm__ : _ Stream.t) = s in
    match Stream.peek strm__ with
      Some (Genlex.Ident x2) ->
        Stream.junk strm__;
        let maps =
          try parse_maps g strm__ with
            Stream.Failure -> raise (Stream.Error "Error 4")
        in
        let r = new Grammar.rule in
        if Array.length maps = 0 then failwith "Empty map function."
        else
          let mt = g#add_mapping_func maps in
          r#set_rhs 1 (g#new_base_category x2); r#set_maptype mt; r
    | _ ->
        match
          try Some (parse_maps g strm__) with
            Stream.Failure -> None
        with
          Some maps ->
            let r = new Grammar.rule in
            if Array.length maps = 0 then failwith "Empty map function."
            else let mt = g#add_mapping_func maps in r#set_maptype mt; r
        | _ -> failwith "Parse error in parse_rhs2"
  with
    Stream.Error _ -> failwith "Stream error in parse_rhs2"
and parse_maps g s =
  try
    let (strm__ : _ Stream.t) = s in
    match Stream.peek strm__ with
      Some (Genlex.Kwd "[") ->
        Stream.junk strm__;
        let m =
          try parse_map g strm__ with
            Stream.Failure -> raise (Stream.Error "")
        in
        let ms =
          try parse_maps g strm__ with
            Stream.Failure -> raise (Stream.Error "Error 6")
        in
        let m2 = Array.make 1 m in Array.append m2 ms
    | _ -> [| |]
  with
    Stream.Error _ -> failwith "Stream error in parse_maps"
and parse_map g s =
  try
    let (strm__ : _ Stream.t) = s in
    match Stream.peek strm__ with
      Some (Genlex.Int x1) ->
        Stream.junk strm__;
        begin match Stream.peek strm__ with
          Some (Genlex.Kwd ",") ->
            Stream.junk strm__;
            begin match Stream.peek strm__ with
              Some (Genlex.Int x2) ->
                Stream.junk strm__;
                let rest =
                  try parse_map2 g strm__ with
                    Stream.Failure -> raise (Stream.Error "Error 8")
                in
                (x1, x2) :: rest
            | _ -> raise (Stream.Error "")
            end
        | _ -> raise (Stream.Error "")
        end
    | _ -> failwith "Parse error in parse_map"
  with
    Stream.Error _ -> failwith "Stream error in parse_map"
and parse_map2 g s =
  try
    let (strm__ : _ Stream.t) = s in
    match Stream.peek strm__ with
      Some (Genlex.Kwd ";") ->
        Stream.junk strm__;
        begin try parse_map g strm__ with
          Stream.Failure -> raise (Stream.Error "Error 9")
        end
    | Some (Genlex.Kwd "]") -> Stream.junk strm__; []
    | _ -> raise Stream.Failure
  with
    Stream.Error _ -> failwith "Stream error in parse_map2"




