(* -*- mode: tuareg                                                        -*-
 ******************************************************************************
 ** $Id: botmcfgtk.ml,v 1.1 2003/03/08 01:05:05 albro Exp $
 ******************************************************************************
 **                                 botmcfg.ml
 **
 **              Copyright (C) 2002, UCLA.  All Rights Reserved.
 ******************************************************************************
 ** Authors:
 **      Daniel M. Albro
 ******************************************************************************
 ** Description:
 **      Bottom-up chart parser for mcfgs.
 ******************************************************************************
 ** History:
 **      10/23/02 - [da]: Initial Creation.
 ******************************************************************************
 ** Bugs:
 ******************************************************************************
 ** To Do:
 *****************************************************************************)

type queue_entry = {
  item : Chart.subitem;
  cat : int
}

exception Success of Chart.chart

let is_terminal c = (c = 0)

let terminal_cat = 0

let set_category_information g ch =
  (** Supply the chart /ch/ with information about categories embodied
    in /g/.  Assumes g#compute_lookup_frequencies() has been called. *)
  for cat = 1 to g#num_cats - 1 do
    ch#set_cat_info cat (g#arity cat) (g#chart_order cat)
  done

let produce_axioms sentence ch agenda g =
  (** Add all of the items in the sentence into the chart as axioms. *)
  for i = 0 to (Array.length sentence) - 1 do
    let (item : Chart.subitem) = [| i; (i + 1); sentence.(i) |] in
      if ch#add terminal_cat item then begin
        (* print_string "Added ";
           Parsetree.print_item terminal_cat item g;
           print_string "\n"; *)
	Queue.add {item = item; cat = terminal_cat} agenda
      end
  done

let produce_terminal_items pat item g ch agenda sl =
  (** Given a rule pattern /pat/ for a terminal rule and an axiomatic
    item /item/, create all items that can be produced by the rule
    described in /pat/ by using /item/.  Add them to the chart
    and the agenda, if they aren't already there. *)
  if item.(2) = pat#get_state 0 then
    let it = [| item.(0); item.(1) |] in
      if ch#add pat#get_out_cat it then begin
        (* print_string "Added ";
           Parsetree.print_item pat#get_out_cat it g;
           print_string "\n"; *)
        if pat#get_out_cat = g#start_cat && it = [|0;sl|] then
          raise (Success ch)
        else
          Queue.add {item=it; cat=pat#get_out_cat} agenda
      end

let produce_nonterminal_items pat item0 g ch agenda sl =
  (** Given a rule pattern /pat/ for a nonterminal rule and a
    non-axiomatic item /item/, create all items that can be produced
    by the rule described in /pat/ by using /item/.  Add them to 
    the chart and the agenda, if they aren't already there. If a new
    item added to the chart is the success item, signal it with
    Success(ch).
  *)
  let item = Array.copy item0 in
    if pat#check_item item then
      let new_items =
        if not (pat#multi_cat_pat()) then
	  [pat#out_item item [||] ]
        else
          let si = pat#search_item item in
            Grammar.separate_variables item si;
            let ms = Parsetree.lookup_matches pat#get_other_cat si ch g sl in
            let rec good_matches its =
              match its with
                  [] -> []
                | it :: tl ->
                    let it' = Array.copy it in
                      if pat#check_other_item it' then
                        it' :: (good_matches tl)
                      else
                        good_matches tl
            in
	    let good_ms = good_matches ms in
            let rec get_outs good_items =
              match good_items with
                  [] -> []
                | it :: tl ->
                    try
                      let (trig, other) = pat#unify_check_items item it in
                        (pat#out_item trig other) :: (get_outs tl)
                    with Grammar.Ununifiable -> get_outs tl
            in
              get_outs good_ms
      in
      let add_item it =
        if ch#add pat#get_out_cat it then
          begin
            (* print_string "Added ";
               Parsetree.print_item pat#get_out_cat it g;
               print_string "\n"; *)
            if pat#get_out_cat = g#start_cat && it = [|0;sl|] then
              raise (Success ch)
            else
              Queue.add {item=it; cat=pat#get_out_cat} agenda
          end
      in
        List.iter add_item new_items

let bottom_up_parse g sentence =
  (** Parse /sentence/ (an array of integers from some symbol table) with
    respect to grammar g.  Return the filled-out chart. *)
  let sl = Array.length sentence in
  let ch = new Chart.chart g#num_cats sl in
  let agenda = Queue.create() in
    (* Tell the chart how to insert/search on each category in g *)
    set_category_information g ch;
    (* Check for empty sentence *)
    if (Array.length sentence) = 0 then
      if g#is_empty_cat g#start_cat then
        raise (Success ch)
      else
        ch
    else
      begin
        (* Add the axiomatic items to the chart/agenda *)
        produce_axioms sentence ch agenda g;
        (* produce_empties ch agenda g sl; *)
        (* Process the agenda *)
        while not (Queue.is_empty agenda) do
          let entry = Queue.take agenda in
          let patterns = g#patterns entry.cat in
          let process_pattern pat =
            if is_terminal entry.cat then
              produce_terminal_items pat entry.item g ch agenda sl
            else
              produce_nonterminal_items pat entry.item g ch agenda sl
          in
            (* print_string "  Trigger is ";
            Parsetree.print_item entry.cat entry.item g;
            print_string "\n"; *)
	    List.iter process_pattern patterns
        done;
        ch
      end

let _ =
  let result = ref 0 in
  let total_time = ref 0.0 in
  let draw_tree = ref false in
  let parse_sentence g str =
    let words = Str.split (Str.regexp "[ \t]+") str in
    let init_time = Unix.gettimeofday() in
    let time_taken =
      try
        let snt = Array.of_list (List.map g#terminal_id words) in
          ignore(bottom_up_parse g snt);
          Printf.printf "Failed to parse %s\n" str;
          result := 1;
          Unix.gettimeofday() -. init_time
      with
          Success ch -> 
            Printf.printf "Parsed %s\n" str;
            let tt = Unix.gettimeofday() -. init_time in
              Printf.printf "Chart size was %d.\n" ch#chart_size;
              let tree = Parsetree.recover_parse_tree g (List.length words) 
                           ch in
                Parsetree.print_parse_tree tree;
                if !draw_tree then
                  Displaytree.display_parse_tree tree;
                result := 0;
                tt
        | Grammar.Unknown_terminal wd -> 
            Printf.printf "Failed to parse %s -- \"%s\" is unknown.\n" str wd;
            result := 2;
            Unix.gettimeofday() -. init_time
    in
      Printf.printf "Process took %f seconds.\n" time_taken;
      total_time := !total_time +. time_taken
  in
  let gf_set = ref false in
  let got_sentence = ref false in
  let g = new Grammar.grammar in
  let handle_arg arg =
    if not !gf_set then begin
      gf_set := true;
      let lexer = Mcfggram.file_lexer arg in
        Mcfggram.parse_grammar g lexer;
        print_string "Checking the grammar...\n";
        g#check_rules();
        print_string "OK.\n";
        print_string "Building pattern database...\n";
        g#build_patterns();
        print_string "Calculating optimal chart lookup order...\n";
        g#compute_lookup_frequencies()
    end 
    else begin
      got_sentence := true;
      parse_sentence g arg
    end
  in
    Arg.parse [("-t", Arg.Set(draw_tree), "draw the parse tree")]
      handle_arg "Usage: botmcfg <grammar-file> [<options>*] [<sentence>*]";
    if not !gf_set then begin
      prerr_string "Usage: botmcfg <grammar-file> [-t] [<sentence>*]\n";
      exit(-1)
    end;
    if not !got_sentence then begin
      try
        while true do
          print_string "Enter sentence to parse, or hit ^D.\n";
          let ln = read_line () in
            parse_sentence g ln
        done
      with End_of_file -> print_string "Bye!\n"
    end;
    Printf.printf "Total time taken was %f seconds.\n" !total_time;
    exit(!result)
    
    

        

