(* -*- mode: tuareg                                                        -*-
******************************************************************************
** $Id: chart.ml,v 1.4 2003/02/12 00:24:40 albro Exp $
******************************************************************************
**                                  chart.ml
**
**              Copyright (C) 2002, UCLA.  All Rights Reserved.
******************************************************************************
** Authors:
**      Daniel M. Albro
******************************************************************************
** Description:
**      The chart for a bottom-up MCFG chart parser; implementation.  
**      Chart items contain a category and 2n states for some n > 0.
******************************************************************************
** History:
**      11/21/02 - [da]: Initial Creation.
******************************************************************************
** Bugs:
******************************************************************************
** To Do:
*****************************************************************************)

type subitem = int array
(** An item on the chart, as presented to the user (sans cat) *)

(* 
 * We represent a chart as an array of maps from integers (1) to maps
 * from integers (2) to maps from integers (3) to integers (4).  What
 * these integers represent depends on the category.  
 *)

(* Declare integers to be ordered types *)
module OrderedInt =
struct
  type t = int
  let compare x1 x2 = x1 - x2
end

(* Map from integers to some other thing... *)
module AssocInt = Map.Make(OrderedInt)

(* Set of integers *)
module IntSet = Set.Make(OrderedInt)

(* The trie is a map of integers to a map of integers to ... to a set *)
type trie = 
    TrieSet of IntSet.t
  | TrieTree of trie AssocInt.t

(* Exception used internally to chart#add *)
exception Not_added

let print_array ar =
  print_string "[|";
  for i = 0 to (Array.length ar) - 1 do
    if i <> 0 then
      print_string "; ";
    print_int ar.(i)
  done;
  print_string "|]"

let useful_item item (sl:int) =
  (** An item is not useful if it has overlapping sentence portions.  Such
    an item can not contribute to a correct answer, I think. *)
  (* This is O(arity^2), as opposed to the commented out one, which is
   * O(sl * arity).  Usually I expect arity^2 is smaller. *)
  let len = Array.length item in
    if len = 3 then
      true
    else
      let ranges_overlap i j k l =
        if k < 0 || l < 0 then
          false
        else if (k <= i && i < l) 
          || (k <= (j-1) && (j-1) < l) 
          || (i <= k && k < j)
          || (i <= (l-1) && (l-1) < j) then
            true
        else
          false
      in
      let rec ui_aux1 i =
        let r0 = item.(i) and r1 = item.(i + 1) in
        let rec ui_aux2 j =
          if j = len then
            true
          else if ranges_overlap r0 r1 item.(j) item.(j + 1) then
              false
          else
            ui_aux2 (j + 2)
        in
          if i = (len - 2) then
            true
          else if r0 >= 0 && r1 >= 0 && not (ui_aux2 (i + 2)) then
            false
          else
            ui_aux1 (i + 2)
      in
        ui_aux1 0

(*
 * Note -- if the grammar arity starts climbing high enough that average
 * sentence length < arity, then this one will be faster.  Otherwise the
 * above is faster.  This one is O(sentence-length * arity)
let useful_item item sl =
  (** An item is not useful if it has overlapping sentence portions.  Such
    an item can not contribute to a correct answer, I think. *)
  let len = Array.length item in
    if len = 3 then
      true
    else
      let i = ref 0 and j = ref 0 and num = ref 0 and useful = ref true in
        i := 0;
        while !i < sl do
          num := 0;
          j := 0;
          while !j < len do
            if item.(!j) > -1 && item.(!j) <= !i && item.(!j + 1) > !i then
              begin
                incr num;
                if !num > 1 then
                  j := len
              end;
            j := !j + 2
          done;
          if !num > 1 then begin 
            i := sl;
            useful := false
          end;
          incr i
        done;
        !useful
*)

(***************
 * Chart Class *
 ***************)
class chart nc sl =
  let order_func i = if i = 0 then [|1;0;2|] else [|1;0|] in
  let tree_depth_func i = if i = 0 then 3 else 2 in
  let num_var_func i = if i = 0 then [|0;0;0|] else [|0;0|] in
object (self)
  
  (************************
   * Private Data Members *
   ************************)

  val mutable tree = Array.init nc (function i -> TrieTree(AssocInt.empty))
  (** Array from cat to state trie.  The type of this is "trie array" *)
                       
  val num_cats = nc                     
  (** The number of categories in the grammar we're parsing *)

  val m_sentence_length = sl
  (** The length of the sentence being parsed *)

  val tree_depth = Array.init nc tree_depth_func
  (** The depth of the tree for each category.  Will have to set this
    outside somewhere.  Defaults to 2 (arity 1) except for the special
    case of category 0, which has the sort of weird depth 3. *)

  val mutable m_order = Array.init nc order_func
  (** The order to search the trie for each category. In general, for
    maximum efficiency, the order should have unused fields first,
    then fields that get searched on from most frequently searched on
    to least frequently.  It's best to have the field with the largest
    number of possible values last.  Thus unused fields go first,
    since these have only one value (0). *)

  val mutable m_has_vars = Array.make nc false
  (** For each category, whether that category has variables (values less
    than -1) stored in its trie.  Searches for variables are more
    expensive, so we don't make them if we don't have to. *)
                             
  val mutable m_num_vars = Array.init nc num_var_func
  (** For each index within each category, the number of possible variables
    that might be there.  This is a number between 0 and the arity of
    the category. *)

  val mutable m_chart_size = 0
  (** Number of items in the chart. *)

  (******************
   * Public Methods *
   ******************)

  method show_cat_info () =
    (** Display the information stored for each category (basically what
      the sort order of the indices for that category is) *)
    for i = 0 to num_cats - 1 do
      Printf.printf "%d) Tree depth: %d, Order: [" i tree_depth.(i);
      for j = 0 to (Array.length m_order.(i)) - 1 do
        if j <> 0 then
          print_string ", ";
        print_int m_order.(i).(j)
      done;
      print_string "]\n"
    done
                        
  method set_cat_info cat arity order =
    (** Set the tree depth and search order for a category.  Assumes
      the category is between 1 and the maximum category. *)
    assert (cat > 0 && cat < num_cats);
    assert (arity > 0 && arity < 10);
    assert ((Array.length order) = (arity * 2));
    tree_depth.(cat) <- arity * 2;
    m_order.(cat) <- order;
    m_num_vars.(cat) <- Array.make (arity * 2) 0

  method chart_size =
    (** Return the number of items in the chart *)
    m_chart_size

  method contains cat item =
    (** Indicate whether the exact item /cat/:/item/ exists in the chart *)
    assert (cat >= 0 && cat < num_cats);
    let rec trie_partial_lookup subtree level =
      let state = m_order.(cat).(level) in
      let key = item.(state) in
        match subtree with
            TrieSet(set) -> IntSet.mem key set
          | TrieTree(st) ->
              try
                let next = AssocInt.find key st in
                  trie_partial_lookup next (level + 1)
              with Not_found -> false
    in
      trie_partial_lookup tree.(cat) 0

  method matches cat item =
    (** Get a list of all items of category /cat/ that match the
      specification of /item/ -- here, anything matches where a field
      in /item/ is -1, and all other values must match exactly, except
      that we can also match chart items that are unspecified in some
      field or fields. *)
    self#matches_aux cat item m_has_vars.(cat)

  method add cat item =
    (** Add an item to the chart, if it isn't already there.  If the item is
      actually added to the chart, return true, else return false *)
    let add_var key state =
      let varnum = ((-1 * key) - 1) in
        m_has_vars.(cat) <- true;
        if varnum > m_num_vars.(cat).(state) then
          m_num_vars.(cat).(state) <- varnum
    in
    let rec trie_insert subtree level =
      assert (level >= 0 && level < tree_depth.(cat));
      let state = m_order.(cat).(level) in
      let key = item.(state) in
        assert (key <> -1);
        match subtree with
            TrieSet(set) ->
              assert (level = (tree_depth.(cat) - 1));
              if IntSet.mem key set then
                raise Not_added
              else begin
                if key < -1 then
                  add_var key state;
                TrieSet(IntSet.add key set)
              end
          | TrieTree(st) ->
              assert (level < (tree_depth.(cat) - 1));
              let next =
                try
                  AssocInt.find key st
                with Not_found -> 
                  if level < (tree_depth.(cat) - 2) then
                    TrieTree(AssocInt.empty)
                  else
                    TrieSet(IntSet.empty)
              in
              let new_next = trie_insert next (level + 1) in
                if key < -1 then
                  add_var key state;
                TrieTree(AssocInt.add key new_next st)
    in
      try
        if not (useful_item item m_sentence_length) then
          false
        else begin
          (* Add the item.  This will raise Not_added if already there. *)
          tree.(cat) <- (trie_insert tree.(cat) 0);
          (* If we get here, there's a new item in the chart. *)
          m_chart_size <- m_chart_size + 1;
          true
        end
      with Not_added -> false

  (*******************
   * Private Methods *
   *******************)

  method private matches_aux cat item match_variables =
    (** Get a list of all items of category /cat/ that match the
      specification of /item/ -- here, anything matches where a field
      in /item/ is -1, and all other values must match exactly,
      unless match_variables is true, in which case a key of >= 0
      value will match values in the chart that are of value < -1.
      These values represent unknowns. *)
    assert (cat >= 0 && cat < num_cats);
    let remap lst =
      let ary = Array.make tree_depth.(cat) 0 in
        for i = 0 to tree_depth.(cat) - 1 do
          (* If m_order.(cat).(i) is j then the ith element should be
           * the jth element *)
          ary.(m_order.(cat).(i)) <- List.nth lst i
        done;
	ary
    in
    let rec trie_partial_search subtree level tail =
      assert ((List.length tail) = level);
      let state = m_order.(cat).(level) in
      let trie_sub_search skey stree slst =
        let lower = trie_partial_search stree (level + 1) (skey::tail) in
          lower @ slst
      in
      let tpsaux key =
        match subtree with
            TrieSet(set) -> 
              if key = -1 then
                List.map (function a -> remap (List.rev (a::tail)))
                  (IntSet.elements set)
              else
                if IntSet.mem key set then
                  [remap (List.rev (key::tail))]
                else
                  []
          | TrieTree(st) ->
              if key = -1 then
                AssocInt.fold trie_sub_search st []
              else
                try
                  let next = AssocInt.find key st in
                    trie_partial_search next (level + 1) (key::tail)
                with Not_found -> []
      in
      let key = item.(state) in
      let results = ref [] in
        (* First look with the actual user-specified key *)
        results := tpsaux key;
        (* Look with all the variables that might be there, if we want to
         * find variables and have some reason to believe they are there. 
         * Note that if key is -1, then we are already getting the variables.
         *)
        if key > -1 && match_variables && m_has_vars.(cat) then
          for i = 1 to m_num_vars.(cat).(state) do
            results := !results @ (tpsaux (-1 * (i + 1)))
          done;
        !results
    in
      trie_partial_search tree.(cat) 0 []
end

