(* -*- mode: tuareg                                                        -*-
******************************************************************************
** $Id: parsetree.ml,v 1.2 2003/02/04 06:57:59 albro Exp $
******************************************************************************
**                                parsetree.ml
**
**              Copyright (C) 2002, UCLA.  All Rights Reserved.
******************************************************************************
** Authors:
**      Daniel M. Albro
******************************************************************************
** Description:
**      Recover the parse tree from the chart used to parse a sentence.
******************************************************************************
** History:
**      12/19/02 - [da]: Initial Creation.
******************************************************************************
** Bugs:
******************************************************************************
** To Do:
*****************************************************************************)

(****************************************************************************
 ****************************** Parse Tree Type *****************************
 ****************************************************************************)

(* What does a parse tree have in it? (1) each subnode should be a list
   of alternatives *)

type parse_cat = 
    ParseTerminal of string
  | ParseNonTerminal of string * int * int array

type parse_rhs = ParseRhs of parse_node list
and parse_node = ParseNode of parse_cat * parse_rhs list

(****************************************************************************
 **************************** Parse Tree Printing ***************************
 ****************************************************************************)

let print_spaces num =
  for i = 1 to num do
    print_string " "
  done

let print_parse_cat cat =
  match cat with
      ParseTerminal(nm) -> Printf.printf "\"%s\" " nm
    | ParseNonTerminal(catstr, catnum, item) ->
        Printf.printf "%s[" catstr;
        for i = 0 to (Array.length item) - 1 do
          if i > 0 then
            print_string ", ";
          print_int item.(i)
        done;
        print_string "] "

let rec print_parse_node nd level =
  let print_parse_node_aux subnode =
    print_parse_node subnode (level + 1)
  in
  let print_parse_rhs rhs =
    match rhs with
        ParseRhs(nodes) ->
          List.iter print_parse_node_aux nodes
  in
  let print_parse_rhs_alt rhs =
    print_parse_rhs rhs;
    print_string " OR "
  in
    match nd with
        ParseNode(cat, chillins) ->
          match cat with
              ParseTerminal(nm) -> Printf.printf "\"%s\"" nm
            | ParseNonTerminal(_,_,_) ->
                print_string "\n";
                print_spaces (3 * level);
                print_string "[";
                print_parse_cat cat;
                if List.length chillins > 0 then
                  ( print_string "--> ";
                    if (List.length chillins) > 1 then
                      List.iter print_parse_rhs_alt chillins
                    else
                      List.iter print_parse_rhs chillins );
                print_string "]"
                  
let print_parse_tree tree =
  print_parse_node tree 0;
  print_string "\n"

(****************************************************************************
 ********************** Stuff Used Here and in Botmcfg **********************
 ****************************************************************************)

let print_item cat subitem g =
  (** Output a human-readable representation of a chart item *)
  let length = 
    if cat > 0 then (Array.length subitem) else (Array.length subitem) - 1
  in
    if cat > 0 then
      Printf.printf "%s(%d)" (g#cat_string cat) cat
    else
      Printf.printf "\"%s\"(%d)" (g#terminal_name subitem.(2)) subitem.(2);
    print_string ":[";
    for i = 0 to length - 1 do
      if i <> 0 then
        print_string ", ";
      print_int subitem.(i)
    done;
    print_string "]"

let lookup_matches cat item ch g sl =
  (** [lookup_matches cat item ch g sl] looks in the chart (sort of) for
    items that match <cat,item>, which can be a template specification.
    Also looks up any empty category items which might match, since they
    aren't actually in the chart. *)
  (* print_string "    Looking for ";
  print_item cat item g;
  print_string "...\n"; *)
  let empties = 
    if g#is_empty_cat cat then
      let out = Array.init (Array.length item) 
                  (function i -> (-1 * (2 + (i / 2)))) in
        try
          [ Grammar.unify out item ]
        with Grammar.Ununifiable -> []
    else
      []
  in
  let nonempties =
    let rec gather_nonempties items =
      match items with
          [] -> []
        | it :: tl -> 
            try
              (Grammar.unify it item) :: (gather_nonempties tl)
            with Grammar.Ununifiable -> (gather_nonempties tl)
    in
      if g#is_nonempty_cat cat then
        gather_nonempties (ch#matches cat item)
      else
        []
  in
  let ms = List.append empties nonempties in
    ms

(****************************************************************************
 **************************** Parse Tree Recovery ***************************
 ****************************************************************************)

let rhs2_key head rhs1 g rpat r =
  match head with
      ParseTerminal(_) -> failwith "error"
    | ParseNonTerminal(headcatstr, headcat, item) ->
        if Grammar.is_terminal r then
          failwith "error"
        else
          let num = 2 * (g#arity r#rhs.(1)) in
          let num0 = 2 * (g#arity r#rhs.(0)) in
          let key = Array.make num (-1) in
            (* First set from head and rhs1 *)
            for i = 0 to (num - 1) do
              let s = rpat#get_rhs_state 1 i in
                if s = -1 then
                  for j = 0 to num0 - 1 do
                    if rpat#is_eqv (0,j) (1,i) then
                      if rhs1.(j) >= 0 then
                        key.(i) <- rhs1.(j)
                  done
                else
                  key.(i) <- item.(s)
            done;
            key

let rhs1_key head g rpat r =
  match head with
      ParseTerminal(_) -> failwith "error"
    | ParseNonTerminal(headcatstr, headcat, item) ->
        if Grammar.is_terminal r then
          [| item.(0); item.(1); r#rhs.(0) |]
        else
          let num = 2 * (g#arity r#rhs.(0)) in
          let key = Array.make num 0 in
            for i = 0 to (num - 1) do
              key.(i) <- let s = rpat#get_rhs_state 0 i in
                if s = -1 then -1 else item.(s)
            done;
            key

(*
    if g#is_terminal rle then
      [ParseRhs([ParseNode(ParseTerminal(g#terminal_name rle#rhs.(0)), [])])]
    else
*)

let get_terminal_child head g ch rle =
  (** [get_terminal_child g ch rle] returns a rhs listing for the terminal
    on the RHS of rule rle (which must be terminal), if that terminal
    is in the chart at the right place. *)
  let id = rle#rhs.(0) in
    match head with
        ParseTerminal(_) -> failwith "error"
      | ParseNonTerminal(headcatstr, headcat, item) ->
          let key = [| item.(0); item.(1); id |] in
            if id = -1 then
              if item.(0) = item.(1) then
                [ParseRhs([ParseNode(ParseTerminal(g#terminal_name id), [])])]
              else
                []
            else
              let matches = ch#matches 0 key in
                if (List.length matches) > 0 then
                  [ParseRhs([ParseNode(ParseTerminal(g#terminal_name id), 
                                       [])])]
                else
                  []

let good_self_match m rpat idx =
  let good_match = ref true in
    for i = 0 to (Array.length m) - 2 do
      for j = i + 1 to (Array.length m) - 1 do
        if (rpat#is_eqv (idx, i) (idx, j)) && (m.(i) <> m.(j)) then
          good_match := false
      done
    done;
    !good_match

let rec get_child head g ch r sl =
  (** [get_child head g ch r] gets all of the alternatives for a child
    of head using rule /r/ *)
  let rpat = g#reconstruction_pattern r in
  let rle = g#get_rule r in
  let key0 = rhs1_key head g rpat rle in
  let good_match0 m0 = good_self_match m0 rpat 0 in
  let good_match1 m1 = good_self_match m1 rpat 1 in
  let matches0 = lookup_matches rle#rhs.(0) key0 ch g sl in
  let alts = ref [] in
  let handle_match0 m0 =
    if rle#length = 1 then
      try 
        rpat#unify_unary m0;
        let subhead0 = ParseNonTerminal(g#cat_string rle#rhs.(0),
                                        rle#rhs.(0), m0) in
          alts := (ParseRhs([ParseNode(subhead0, 
                                       get_children subhead0 g ch sl)]) 
                   :: !alts)
      with Grammar.Ununifiable -> ()
    else
      let key1 = rhs2_key head m0 g rpat rle in
      let matches1 = lookup_matches rle#rhs.(1) key1 ch g sl in
      let handle_match1 m1 =
        try
          let m0' = rpat#unify_binary m0 m1 in
          let subhead0 = ParseNonTerminal(g#cat_string rle#rhs.(0),
                                          rle#rhs.(0), m0') in
          let subhead1 = ParseNonTerminal(g#cat_string rle#rhs.(1),
                                          rle#rhs.(1), m1) in
            alts := (ParseRhs([ParseNode(subhead0, 
                                         get_children subhead0 g ch sl);
                               ParseNode(subhead1, 
                                         get_children subhead1 g ch sl)
                              ]) :: !alts)
        with Grammar.Ununifiable -> ()
      in
        List.iter handle_match1 matches1
  in
    List.iter handle_match0 matches0;
    !alts
and get_children head g ch sl =
  (** [get_children head g ch] Get all alternative forms (as a list)
    derivable from head, which is a parse_cat. *)
  match head with
      ParseTerminal(term) -> []
    | ParseNonTerminal(catstr, catnum, item) ->
        let rs = g#rules_headed_by catnum in
        let rec get_alts alts rules =
          match rules with
              r :: tail ->
                let rle = g#get_rule r in
                let child = 
                  if Grammar.is_terminal rle then
                    get_terminal_child head g ch rle
                  else
                    get_child head g ch r sl
                in
                  get_alts (alts @ child) tail
            | _ -> alts
        in
          get_alts [] rs

let recover_parse_tree g sl ch =
  (** [recover_parse_tree g sl ch] recovers the parse tree for a successful
    parse of a sentence of length /sl/ using grammar /g/, where the parse
    produced the chart /ch/.  If the parse was not actually successful, 
    raises Not_found.  Otherwise, returns a parse node. *)

  (* 1) Get matches for g#start_cat 0,sl *)
  let starts = lookup_matches g#start_cat [|0;sl|] ch g sl in
  let start_item = 
    try
      List.hd starts
    with
        Failure _ -> raise Not_found
  in
  let start_cat = ParseNonTerminal("S", g#start_cat, [|0;sl|]) in
    ParseNode(start_cat, get_children start_cat g ch sl)
              


