(* -*- mode: tuareg                                                        -*-
******************************************************************************
** $Id: tree.ml,v 1.2 2003/03/11 01:14:56 albro Exp $
******************************************************************************
**                                  tree.ml
**
**              Copyright (C) 2003, UCLA.  All Rights Reserved.
******************************************************************************
** Authors:
**      Daniel M. Albro
******************************************************************************
** Description:
**
******************************************************************************
** History:
**      02/28/03 - [da]: Initial Creation.
******************************************************************************
** Bugs:
******************************************************************************
** To Do:
*****************************************************************************)

open Tk
open Widget

type pref_type = { 
  mutable xgap : int;
  mutable ygap : int;
  mutable triwidth : int;
  mutable xslop : int;
  mutable yslop : int
}

(* Default preferences *)
let prefs = { xgap=10; ygap=10; triwidth=30; xslop=2; yslop=2 }

type tree_node_type = RectTreeNode | TextTreeNode

type side_point = { mutable x : int; mutable y : int }

(** Useful anywhere auxilliaries *)

let rec last = function
    [] -> failwith "last"
  | hd::[] -> hd
  | hd::tl -> last tl

let gensym, init_gensym =
  let ctr = ref 0 in
    (fun tag -> tag ^ (string_of_int !ctr)),
    (fun () -> ctr := 0)

(** Auxilliaries *)
let translate_sides sides dx dy =
  let translate side =
    side.x <- side.x + dx;
    side.y <- side.y + dy
  in
    List.iter translate sides

let extreme_sides op ps qs =
  (** For each y value in ps and qs return (x, y) such that
    x is the one of which (x<ps op x<qs) holds true. 
    Assumes both are sorted by y value. *)
  let rec aux plst qlst =
    match plst, qlst with
        [], q -> q
      | p, [] -> p
      | ({x=px; y=py} as p)::ptl, ({x=qx; y=qy} as q)::qtl ->
          if py < qy then
            p::(aux ptl qlst)
          else if py = qy then
            if op px qx then
              p::(aux ptl qtl)
            else
              q::(aux ptl qtl)
          else
            q::(aux plst qtl)
  in
    aux ps qs


let max_overlap rights lefts =
  (** Get the largest distance between left and right at any point where left
    and right are at the same y value *)
  let rec aux max_xgap rs ls =
    match rs, ls with
        [], _ | _, [] -> max_xgap
      | {x=rx; y=ry}::rtail, {x=lx; y=ly}::ltail ->
          if ry = ly then
            let xgap = rx - lx in
              aux (if xgap > max_xgap then xgap else max_xgap) rtail ltail
          else if ry < ly then
            aux max_xgap rtail ls
          else
            aux max_xgap rs ltail
  in
    aux 0 rights lefts


let layout_subtrees dc trees dy left_side right_side =
  let y_max = ref 0 in
  let left_side1 = ref [] in
  let right_side1 = ref [] in
  let iter tree =
    tree#layout dc left_side1 right_side1;
    let dx =
      if (List.length !right_side) = 0 then
        0
      else
        (max_overlap !right_side !left_side1) + prefs.xgap
    in
      if tree#height + dy > !y_max then
        y_max := tree#height + dy;
      tree#root#set_delta dx dy;
      translate_sides !right_side1 dx 0;
      right_side := extreme_sides (>) !right_side1 !right_side;
      right_side1 := [];
      translate_sides !left_side1 dx 0;
      left_side := extreme_sides (<) !left_side1 !left_side;
      left_side1 := [];
      tree#translate_minmax dx
  in
    left_side := [];
    right_side := [];
    List.iter iter trees;
    !y_max

let min_x_side sides =
  let rec aux min_x = function
      [] -> min_x
    | {x=sx; y=_}::tl ->
        aux (if sx < min_x then sx else min_x) tl
  in
    aux 0 sides

let max_x_side sides =
  let rec aux max_x = function
      [] -> max_x
    | {x=sx; y=_}::tl ->
        aux (if sx < max_x then sx else max_x) tl
  in
    aux 0 sides

let print_tag tag =
  match tag with
      `Id(num) -> Printf.printf "Id(%d)" num
    | `Tag(str) -> Printf.printf "Tag(%s)" str

let clipped_line x1 x2 y1 y2 x_min x_max y_top y_bot =
  (** Given a line from (x1, y1) down to (x2, y2), return the line reduced
    such that it doesn't go higher than y_tap or lower than y_bot. *)
  assert (y1 < y2 && y_top < y_bot);

  (* 1 - Find the intersection point of the line with y = y_top *)
  if (x2 - x1) = 0 then
    (x1, x2, y_top, y_bot)
  else
    let m = ((float_of_int y2) -. (float_of_int y1)) /. 
              ((float_of_int x2) -. (float_of_int x1)) in
    let b = ((float_of_int y1) -. (m *. (float_of_int x1))) in
    let new_x1 = (int_of_float (((float_of_int y_top) -. b) /. m)) in
    let new_x2 = (int_of_float (((float_of_int y_bot) -. b) /. m)) in
    let new_y1 = (int_of_float ((m *. (float_of_int x_min)) +. b)) in
    let new_y2 = (int_of_float ((m *. (float_of_int x_max)) +. b)) in
      ((min new_x1 x_min), (max new_x2 x_max), 
       (min new_y1 y_top), (max new_y2 y_bot))

class virtual tree_node =
object (self)
  val mutable m_width = -1
  val mutable m_height = -1
  val mutable m_dx = -1
  val mutable m_dy = -1
  val mutable m_xchildshift = -1
  val mutable m_type = TextTreeNode
  val mutable m_tags = []

  method is_drop_site () = false
  method is_simple_node () = false
  method is_collapsed () = false
  method get_node_size dc = self#recalc_node_size dc; (m_width, m_height)
  method virtual draw : canvas widget -> int -> int -> unit
  method drag dc dx dy =
    List.iter (fun tag -> Canvas.move dc tag ~x:dx ~y:dy) m_tags    
  method set_delta dx dy = m_dx <- dx; m_dy <- dy
  method dx = m_dx
  method dy = m_dy
  method xchildshift = m_xchildshift
  method set_xchildshift xcs = m_xchildshift <- xcs
  method width = m_width
  method height = m_height
  method node_type = m_type
  method collapse (dc : canvas widget) = ()
  method expand (dc : canvas widget) = ()
  method virtual recalc_node_size : canvas widget -> unit
  method associate_tag dc tag =
    let iter t = Canvas.addtag dc ~tag ~specs:[`Withtag(t)]
    in
      List.iter iter m_tags
  method get_tag = (List.hd m_tags)
end

class rect_tree_node width height =
object (self)
  inherit tree_node as super

  method is_drop_site () = true

  method draw dc x y =
    let w2 = m_width / 2 in
    let h2 = m_height / 2 in
    let p1x = x - w2 and p1y = y - h2 in
    let p2x = x - w2 + m_width and p2y = y - h2 in
    let p3x = x - w2 + m_width and p3y = y + h2 in
    let p4x = x - w2 and p4y = y + h2 in
    let tag = Canvas.create_polygon dc ~outline:`Black ~width:1 ~fill:`White
                ~xys:[(p1x, p1y); (p2x, p2y); (p3x, p3y); (p4x, p4y)] in
      m_tags <- [tag]
      
  method recalc_node_size dc = ()
    
  initializer
    m_width <- width;
    m_height <- height;
    m_type <- RectTreeNode
end

class text_tree_node text =
object (self)
  inherit tree_node as super
  val mutable m_text = text
  val mutable m_text_width = -1
  val mutable m_text_height = -1
  val mutable m_is_collapsed = false

  method is_collapsed () = m_is_collapsed
  method is_simple_node () = true
  method draw dc x y =
    let tag = Canvas.create_text ~x ~y ~text:m_text dc in
      m_tags <- [tag];
      self#recalc_node_size dc
  method collapse dc = m_is_collapsed <- true; self#recalc_node_size dc
  method expand dc = m_is_collapsed <- false; self#recalc_node_size dc
  method text = m_text
  method set_text dc text = m_text <- text; self#recalc_node_size dc
  method recalc_node_size dc =
    match m_tags with
        [] -> 
          let tag = Canvas.create_text ~x:0 ~y:0 ~text:m_text dc in
          let x1, y1, x2, y2 = Canvas.bbox dc [tag] in
            m_text_width <- (x2 - x1);
            m_text_height <- (y2 - y1);
            m_width <- m_text_width;
            m_height <- m_text_height;
            Canvas.delete dc [tag]
      | _ ->
          let x1, y1, x2, y2 = Canvas.bbox dc m_tags in
            m_text_width <- (x2 - x1);
            m_text_height <- (y2 - y1);
            m_width <- m_text_width;
            m_height <- m_text_height

  initializer
    m_type <- TextTreeNode
end

(** Tree class *)
class tree (root : tree_node) =
object (self)
  val mutable m_root = root
  val mutable m_children = []
  val mutable m_collapsed_children = []
  val mutable m_big_dx = 0
  val mutable m_big_dy = 0
  val mutable m_x_min = 0
  val mutable m_x_max = 0
  val mutable m_y_max = 0
  val mutable m_x = 0
  val mutable m_y = 0
  val mutable m_width = 0
  val mutable m_height = 0
  val mutable m_tags = []               (* Tk tags.  Delete these do redraw. *)
  val mutable m_last_drawn_x = None
  val mutable m_last_drawn_y = None

  (* Accessors *)
  method root = m_root
  method x = m_x
  method y = m_y
  method width = m_width
  method height = m_height

  method get_tag = m_root#get_tag
  method associate_tag dc tag =
    let iter t = Canvas.addtag dc ~tag ~specs:[`Withtag(t)]
    in
      List.iter iter m_tags;
      m_root#associate_tag dc tag;
      List.iter (fun child -> child#associate_tag dc tag) m_children

  method add_child (t:tree) = m_children <- m_children @ [t]

  method remove_child idx =
    let rec aux i lst =
      match lst with
          [] -> []
        | hd::tl ->
            if i = idx then tl else hd::(aux (i + 1) tl)
    in
      m_children <- aux 0 m_children

  method has_children () = (List.length m_children) <> 0

  method expand dc =
    m_root#expand dc;
    let t = m_children in
      m_children <- m_collapsed_children;
      m_collapsed_children <- t

  method collapse dc =
    if self#has_children() then begin
      m_root#collapse dc;
      let t = m_children in
        m_children <- m_collapsed_children;
        m_collapsed_children <- t
    end

  method replace_child idx child =
    let rec aux i lst =
      match lst with
          [] -> []
        | hd::tl ->
            if i = idx then child::tl else hd::(aux (i + 1) tl)
    in
      m_children <- aux 0 m_children

  method cnp_aux dc px py xcs bpx bpy =
    let aux child = child#cnp_aux dc m_x m_y m_root#xchildshift bpx bpy in
      m_x <- px + m_root#dx + xcs;
      m_y <- py + m_root#dy;
      m_big_dx <- bpx - m_x;
      m_big_dy <- bpy - m_y;
      List.iter aux m_children

  method compute_node_positions (dc : canvas widget) =
    let aux child = child#cnp_aux dc m_x m_y m_root#xchildshift m_x m_y in
      List.iter aux m_children

  method was_hit dc x y =
    let w, h = m_root#get_node_size dc in
    let w2 = w / 2 in
    let top_x = m_x - w2 - prefs.xslop in
    let top_y = m_y - prefs.yslop in
    let bot_x = m_x + w2 + prefs.xslop in
    let bot_y = m_y + h + prefs.yslop in
      (x >= top_x && x <= bot_x && y >= top_y && y <= bot_y)
    
  method find_hitee dc x y (parent : tree) (cn : int) =
    (** @return Some(hitee, hitee's parent, child # hitee is of parent),
      or None. *)
    if self#was_hit dc x y then
      Some((self :> tree), parent, cn)
    else
      let rec aux i (lst : tree list) =
        match lst with
            [] -> None
          | child::tl -> 
              let hitee = (child#find_hitee dc x y (self :> tree) i) in
                match hitee with
                    None -> aux (i + 1) tl
                  | Some(t, tp, tcn) -> hitee
      in
        aux 0 m_children

  method translate_minmax dx =
    m_x_min <- m_x_min + dx;
    m_x_max <- m_x_max + dx;
    List.iter (fun child -> child#translate_minmax dx) m_children

  method min_x =
    let cur_min_x = ref (m_x - (m_root#width / 2)) in
    let update_min child =
      let child_min_x = child#min_x in
        if child_min_x < !cur_min_x then
          cur_min_x := child_min_x
    in
      List.iter update_min m_children;
      !cur_min_x

  method max_x =
    let cur_max_x = ref (m_x + (m_root#width / 2)) in
    let update_max child =
      let child_max_x = child#max_x in
        if child_max_x > !cur_max_x then
          cur_max_x := child_max_x
    in
      List.iter update_max m_children;
      !cur_max_x

  method min_y =
    let cur_min_y = ref (m_y - (m_root#height / 2)) in
    let update_min child =
      let child_min_y = child#min_y in
        if child_min_y < !cur_min_y then
          cur_min_y := child_min_y
    in
      List.iter update_min m_children;
      !cur_min_y

  method max_y =
    let cur_max_y = ref (m_y + (m_root#height / 2)) in
    let update_max child =
      let child_max_y = child#max_y in
        if child_max_y > !cur_max_y then
          cur_max_y := child_max_y
    in
      List.iter update_max m_children;
      !cur_max_y

  method calculate_size () =
    m_width <- self#max_x - self#min_x;
    m_height <- self#max_y - self#min_y

  method layout dc left_side right_side =
    let w, h = m_root#get_node_size dc in
    let w2 = ref (w / 2) in
      left_side := [];
      right_side := [];
      m_y_max <- h;
      if self#has_children() then begin
        let dy = h + prefs.ygap in
        let x = (last m_children)#root#dx in
        let dx = x / (-2) in
          m_y_max <- layout_subtrees dc m_children dy left_side right_side;
          m_root#set_xchildshift dx;
          translate_sides !left_side dx dy;
          translate_sides !right_side dx dy;
          List.iter (fun child -> child#translate_minmax dx) m_children
      end;
      if self#collapsed then begin
        let total_h = h + prefs.triwidth in
          if !w2 < (prefs.triwidth / 2) then
            w2 := (prefs.triwidth / 2);
          if m_y_max < h then
            m_y_max <- h
      end;
      left_side := {x=(-(!w2)); y=0} :: !left_side;
      right_side := {x=(!w2); y = 0} :: !right_side;
      m_x_min <- min_x_side !left_side;
      m_x_max <- max_x_side !right_side;
      m_width <- m_x_max - m_x_min;
      m_height <- m_y_max
        
  method set_location x y =
    m_x <- x;
    m_y <- y;
    m_big_dx <- 0;
    m_big_dy <- 0

  method erase dc =
    Canvas.delete dc m_tags;
    m_tags <- [];
    m_last_drawn_x <- None;
    m_last_drawn_y <- None;
    List.iter (fun child -> child#erase dc) m_children

  method collapsed = (List.length m_collapsed_children) <> 0

  method drag dc =
    let w, h = m_root#get_node_size dc in
    let x = m_x in
    let y = m_y + (h / 2) in
      match m_last_drawn_x, m_last_drawn_y with
          Some(lx), Some(ly) ->
            let dx = x - lx in
            let dy = y - lx in
              List.iter (fun tag -> Canvas.move dc tag ~x:dx ~y:dy) m_tags;
              List.iter (fun child -> child#drag dc) m_children;
              m_root#drag dc dx dy;
              m_last_drawn_x <- Some(x);
              m_last_drawn_y <- Some(y)
        | _, _ -> self#erase dc; self#draw dc

  method draw dc =
    m_root#draw dc m_x m_y;
    List.iter (fun child -> child#draw dc) m_children;
    let w, h = m_root#get_node_size dc in
    let x = m_x in
    let y = m_y + (h / 2) in
    let xmin = x + (w / 2) + 2 in
    let rw = ref w in
    let rh = h - 2 in
      m_tags <- [];
      if not self#collapsed then begin
        let draw_child child =
          let x1 = child#x in
          let w, h = child#root#get_node_size dc in
          let y1 = child#y - (h / 2) 
                   - (if child#root#node_type = RectTreeNode then 2 else 0) 
          in
          let xmax = x1 - (w / 2) - 2 in
          let x', x1', y', y1' = 
            clipped_line x x1 m_y child#y xmin xmax y y1 in
          let tag = Canvas.create_line ~xys:[(x', y'); (x1', y1')] dc in
            m_tags <- tag :: m_tags
        in
          List.iter draw_child m_children
      end else begin
        let triwidth2 = prefs.triwidth / 2 in
        let triheight = prefs.triwidth in
        let p1x = m_x in
        let p1y = m_y + m_root#height in
        let p2x = m_x - triwidth2 in
        let p2y = p1y + triheight in
        let p3x = m_x + triwidth2 in
        let p3y = p2y in
        let tag = Canvas.create_polygon ~fill:`White ~outline:`Black ~width:1
                    ~xys:[(p1x, p1y); (p2x, p2y); (p3x, p3y)] dc in
          m_tags <- tag :: m_tags
      end
end (* tree *)


    
